
/****************************************************************
 *
 *	rtp_amr.c		(C) Copyright Netgem 2008
 *	RTP AMR demuxer
 *
 ****************************************************************/

#include "rtp_amr.h"
#include "libavcodec/bitstream.h"
#include "libavutil/avstring.h"

//#define DEBUG

struct PayloadContext {
    int octet_aligned;
    int interleaving;
};

static PayloadContext *amr_new_extradata(void)
{
    return av_mallocz(sizeof(PayloadContext));
}

static void amr_free_extradata(PayloadContext *d)
{
    av_free(d);
}

static void sdp_parse_fmtp_config_amr(AVStream * stream,
                                      PayloadContext * data,
                                      char *attr, char *value)
{
    if (!strcmp(attr, "octet-align")) {
        if (value[0] != '\0')
            data->octet_aligned = atoi(value);
        else
            data->octet_aligned = 1;
    }
    else if (!strcmp(attr, "interleaving")) {
        data->interleaving = atoi(value);
    }
}

static int parse_amr_sdp_line(AVStream * stream, PayloadContext *data,
                               const char *line)
{
    const char *p = line;

    if (av_strstart(p, "fmtp:", &p)) {
        char attr[256];
        char value[4096];

        while (*p && *p == ' ') p++; // strip spaces.
        while (*p && *p != ' ') p++; // eat protocol identifier
        while (*p && *p == ' ') p++; // strip trailing spaces.

        /* loop on each attribute */
        while (rtsp_next_attr_and_value
               (&p, attr, sizeof(attr), value, sizeof(value))) {
            /* grab the codec extra_data from the config parameter of the fmtp line */
            sdp_parse_fmtp_config_amr(stream, data, attr, value);
        }
    }

    return 0;
}
  
static int amr_handle_packet(PayloadContext *data,
                             AVStream *st,
                             AVPacket * pkt,
                             uint32_t * timestamp,
                             const uint8_t * buf,
                             int len, int flags)
{
    int err;
  
    if (!data->octet_aligned) {
        av_log(NULL, AV_LOG_ERROR, "bandwidth-efficient mode not supported\n");
        return -1;
    }
    if (data->interleaving) {
        av_log(NULL, AV_LOG_ERROR, "interleaving not supported\n");
        return -1;
    }

    if (len < 1)
        return -1;

    /* 4 bits for "codec mode request" + 4 bits reserved */
    buf++;
    len--;

    dprintf(NULL, "amr_handle_packet len=%d\n", len);

    if (!len)
        return -1;

    err = av_new_packet(pkt, len);
    if (err < 0)
        return -1;
    memcpy(pkt->data, buf, len);

    return 0;
}

/**
This is the structure for expanding on the dynamic rtp protocols (makes everything static. yay!)
*/
RTPDynamicProtocolHandler ff_amrnb_dynamic_handler = {
    "AMR",
    CODEC_TYPE_AUDIO,
    CODEC_ID_AMR_NB,
    parse_amr_sdp_line,
    amr_new_extradata,
    amr_free_extradata,
    amr_handle_packet
};

RTPDynamicProtocolHandler ff_amrwb_dynamic_handler = {
    "AMR-WB",
    CODEC_TYPE_AUDIO,
    CODEC_ID_AMR_WB,
    parse_amr_sdp_line,
    amr_new_extradata,
    amr_free_extradata,
    amr_handle_packet
};
