/*
 * modprobe
 */
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <elf.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>

#include <libng/unistd.h>
#include <libng/string.h>

#include "util.h"
#include "module.h"

typedef struct Module {
    struct Module *next;
    char *filename;
    int nb_deps;
    char **deps;
    int explored;
    int cycled;
} Module;

typedef struct Alias {
    char *name;
    char *value;
} Alias;

typedef struct ModuleOptions {
    char *name;
    int nb_options;
    char **options;
} ModuleOptions;

static Module *first_module;

static Alias **aliases;
static int nb_aliases;

static ModuleOptions **module_options;
static int nb_module_options;

const char *modules_paths[] = {
    "/lib/modules",
    "/usr/lib/modules",
    "/home/lib/modules",
    NULL,
};

static int verbose;
static char *loaded_modules;
static int nb_loaded_modules;

/*
 *	Strip white char at the end of a string.
 *	Return the address of the last non white char + 1 (point on the '\0').
 */
static char *strip_end(char *str)
{
	int len = strlen(str);

	for (str += len - 1; len > 0 && (isspace(*str)); --len, --str)
		*str = '\0';
	return str + 1;
}

/*
 *	Read a line of a configuration file and process continuation lines.
 *	Return buf, or NULL if EOF.
 *	Blank at the end of line are always stripped.
 *	Everything on a line following comchar is a comment.
 *
 *	Continuation character is \
 *	Comment character is #
 */
char *fgets_strip(char *buf, int sizebuf, FILE * fin, int *lineno)
{
	int nocomment = 1; /* No comments found ? */
	int contline = 0;
	char *start = buf;
	char *ret = NULL;
	char comchar = '#';
	char contchar = '\\';

	*buf = '\0';

	while (fgets(buf, sizebuf, fin) != NULL) {
		char *end = strip_end(buf);
		char *pt = vstrchr(buf, comchar);

		if (pt != NULL) {
			nocomment = 0;
			*pt = '\0';
			end = strip_end(buf);
		}

		if (lineno != NULL)
			(*lineno)++;
		ret = start;
		if (contline) {
			char *pt = buf;

			while (isspace(*pt))
				pt++;
			if (pt > buf + 1) {
				strcpy(buf + 1, pt);	/* safe, backward copy */
				buf[0] = ' ';
				end -= (int) (pt - buf) - 1;
			} else if (pt == buf + 1) {
				buf[0] = ' ';
			}
		}
		if (end > buf && *(end - 1) == contchar) {
			if (end == buf + 1 || *(end - 2) != contchar) {
				/* Continuation */
				contline = 1;
				end--;
				*end = '\0';
				buf = end;
			} else {
				*(end - 1) = '\0';
				break;
			}
		} else {
			break;
		}
	}

	return ret;
}

/* get a word delimited with space or ending with 'sep' */
static void getword(char *buf, int buf_size, const char *sep, const char **pp)
{
    const char *p;
    char *q;

    p = *pp;
    while (isspace(*p))
        p++;
    q = buf;
    while (*p != '\0' && (!sep || !strchr(sep, *p)) && !isspace(*p)) {
        if ((q - buf) < buf_size - 1)
            *q++ = *p;
        p++;
    }
    *q = '\0';
    *pp = p;
}

static int read_depfile(const char *filename)
{
    FILE *f;
    char line[1024];
    char modname[1024];
    const char *p;
    Module *m;

    f = fopen(filename, "r");
    if (!f)
        return -1;
    for(;;) {
        if (!fgets_strip(line, sizeof(line), f, NULL))
            break;
        /* read module name */
        p = line;
        getword(modname, sizeof(modname), ":", &p);
        while (isspace(*p))
            p++;
        if (*p != ':')
            continue;
        p++;
        
        /* read dependancy list */
        m = xmalloc(sizeof(Module));
        memset(m, 0, sizeof(Module));
        m->filename = xstrdup(modname);
        for(;;) {
            getword(modname, sizeof(modname), NULL, &p);
            if (modname[0] == '\0')
                break;
            dynarray_add(&m->deps, &m->nb_deps, xstrdup(modname));
        }
        /* add the module */
        m->next = first_module;
        first_module = m;
    }
    fclose(f);
    return 0;
}

static int read_modules_conf(const char *filename)
{
    FILE *f;
    char line[1024];
    char cmd[64];
    const char *p;
    int is_add, i;
    
    f = fopen(filename, "r");
    if (!f)
        return -1;
    for(;;) {
        if (!fgets_strip(line, sizeof(line), f, NULL))
            break;
        p = line;
        getword(cmd, sizeof(cmd), NULL, &p);
        if (!strcmp(cmd, "options")) {
            /* add options for a given module */
            ModuleOptions *opt;
            char name[256];
            char option[1024];

            is_add = 0;
        do_options:
            getword(name, sizeof(name), NULL, &p);
            /* find the option */
            for(i = 0; i < nb_module_options; i++) {
                opt = module_options[i];
                if (!strcmp(opt->name, name))
                    goto option_found;
            }
            opt = xmalloc(sizeof(ModuleOptions));
            memset(opt, 0, sizeof(ModuleOptions));
            opt->name = xstrdup(name);
        option_found:
            if (!is_add) {
                free(opt->options);
                opt->options = NULL;
                opt->nb_options = 0;
            }
            for(;;) {
                getword(option, sizeof(option), NULL, &p);
                if (option[0] == '\0')
                    break;
                dynarray_add(&opt->options, &opt->nb_options, xstrdup(option));
            }
            dynarray_add(&module_options, &nb_module_options, opt);
        } else if (!strcmp(cmd, "add")) {
            getword(cmd, sizeof(cmd), NULL, &p);
            if (strcmp(cmd, "options") != 0)
                error("options keyword expected");
            is_add = 1;
            goto do_options;
        } else if (!strcmp(cmd, "alias")) {
            Alias *al;
            char name[256];
            char value[256];
            getword(name, sizeof(name), NULL, &p);
            getword(value, sizeof(value), NULL, &p);
            al = xmalloc(sizeof(Alias));
            al->name = xstrdup(name);
            al->value = strdup(value);
            dynarray_add(&aliases, &nb_aliases, al);
        } else {
            error("invalid command %s", cmd);
        }
    }
    fclose(f);
    return 0;
}

static void dump_deps(void)
{
    Module *m;
    int i;

    for(m = first_module; m != NULL; m = m->next) {
        printf("%s:", m->filename);
        for(i=0;i<m->nb_deps;i++)
            printf(" %s", m->deps[i]);
        printf("\n");
    }
}

static Module **modules_to_load;
static int nb_modules_to_load;

static void add_module(const char *filename)
{
    int i;
    Module *m;

    for(m = first_module ; m != NULL; m = m->next) {
        if (!strcmp(m->filename, filename))
            break;
    }
    if (!m)
        error("module '%s' not found", filename);
    if (m->cycled)
        error("cycle detected in dependancies");
    if (m->explored)
        return;
    m->cycled = 1;
    m->explored = 1;
    /* add the dependancies first */
    for(i=0;i<m->nb_deps;i++)
        add_module(m->deps[i]);
    /* then the module itself */
    dynarray_add(&modules_to_load, &nb_modules_to_load, m);
    m->cycled = 0;
}

static void shortname(char *buf, int buf_size, const char *filename)
{
    const char *p, *q;
    int len;

    p = strrchr(filename, '/');
    if (!p)
        p = filename;
    else
        p++;
    q = strrchr(p, '.');
    if (!q)
        q = p + strlen(p);
    len = q - p;
    if (len > buf_size - 1)
        len = buf_size - 1;
    memcpy(buf, p, len);
    buf[len] = '\0';
}

static Module *find_module(const char *m_name)
{
    char buf[1024];
    Module *m;

    for(m = first_module ; m != NULL; m = m->next) {
        shortname(buf, sizeof(buf), m->filename);
        if (!strcmp(buf, m_name))
            return m;
    }
    return NULL;
}

static void get_loaded_modules(void)
{
    int loaded_modules_size;

    loaded_modules_size = 256;
    loaded_modules = xmalloc(loaded_modules_size);
    for(;;) {
        if (query_module(NULL, QM_MODULES, 
                         loaded_modules, loaded_modules_size, 
                         &nb_loaded_modules) == 0)
            break;
        if (errno != ENOSPC)
            error("QM_MODULES: %m");
        loaded_modules = xrealloc(loaded_modules, nb_loaded_modules);
    }
}

static int is_loaded_module(const char *m_name)
{
    const char *p;
    int i;

    p = loaded_modules;
    for(i = 0; i < nb_loaded_modules; i++) {
        if (!strcmp(m_name, p))
            return 1;
        p += strlen(p) + 1;
    }
    return 0;
}

static int find_module_path(char *path, int path_size, const char *filename)
{
    struct stat st;
    const char **pp;
    const char *b_name;
    
    if (stat(filename, &st) == 0) {
        pstrcpy(path, path_size, filename);
        return 0;
    }
    b_name = basename(filename);
    for(pp = modules_paths; *pp != NULL; pp++) {
        snprintf(path, path_size, "%s/%s", *pp, b_name);
        if (stat(path, &st) == 0) {
            return 0;
        }
    }
    return -1;
}

static void load_modules(Module **modules_to_load, int nb_modules_to_load)
{
    Module *m;
    char m_name[1024];
    char filename[1024];
    char **argv;
    int ret, i, j, k, nb_argv;

    for(i = 0;i < nb_modules_to_load; i++) {
        m = modules_to_load[i];
        shortname(m_name, sizeof(m_name), m->filename);
        if (is_loaded_module(m_name))
            continue;
        if (find_module_path(filename, sizeof(filename), m->filename) < 0) {
            fprintf(stderr, "Could not find module '%s'\n", m->filename);
            exit(1);
        }
        if (verbose) {
            printf("loading '%s'\n", filename);
        }
        argv = NULL;
        nb_argv = 0;
        dynarray_add(&argv, &nb_argv, (char *)"insmod");
        dynarray_add(&argv, &nb_argv, (char *)filename);
        /* adds the option if they are specified */
        for(j = 0; j < nb_module_options; j++) {
            ModuleOptions *opt = module_options[j];
            if (!strcmp(opt->name, m_name)) {
                for(k = 0; k < opt->nb_options; k++) {
                    dynarray_add(&argv, &nb_argv, opt->options[k]);
                    printf("option '%s'\n", opt->options[k]);
                }
                break;
            }
        }
        dynarray_add(&argv, &nb_argv, (char *)NULL);
        ret = insmod_main(nb_argv - 1, argv);
        if (ret)
            exit(ret);
    }
}

static void remove_modules(Module **modules_to_remove, 
                           int nb_modules_to_remove)
{
    Module *m;
    char m_name[1024];
    int i;

    for(i = nb_modules_to_remove - 1; i >=0; i--) {
        m = modules_to_remove[i];
        shortname(m_name, sizeof(m_name), m->filename);
        delete_module(m_name);
    }
}

static void usage(void)
{
    printf("usage: modprobe [-h] [-d] [-v] [-r] module_name\n");
    exit(1);
}

int modprobe_main(int argc, char **argv)
{
    const char **pp;
    int c, opt_dump, i;
    char buf[1024];
    const char *m_name;
    Module *m;
    int remove;
    const char *modules_conf;
    
    opt_dump = 0;
    remove = 0;
    modules_conf = "/etc/modules.conf";
    while ((c = getopt(argc, argv, "hdvrC:")) != EOF) {
        switch (c) {
        case 'd':
            opt_dump = 1;
            break;
        case 'v':
            verbose = 1;
            break;
        case 'r':
            remove = 1;
            break;
        case 'C':
            modules_conf = optarg;
            break;
        default:
            usage();
            break;
        }
    }

    /* load all dependency files */

    for(pp = modules_paths; *pp != NULL; pp++) {
        snprintf(buf, sizeof(buf), "%s/modules.dep", *pp);
        read_depfile(buf);
    }

    read_modules_conf(modules_conf);

    if (opt_dump) {
        dump_deps();
        exit(0);
    }

    if (optind >= argc)
        usage();
    
    m_name = argv[optind];

    /* use the alias table (single level) */
    {
        int i;
        for(i=0;i<nb_aliases;i++) {
            if (!strcmp(aliases[i]->name, m_name)) {
                m_name = aliases[i]->value;
                break;
            }
        }
    }

    /* find module */
    m = find_module(m_name);
    if (!m) {
        fprintf(stderr, "Could not find module '%s'\n", m_name);
        exit(1);
    }
    
    /* find all dependancies */
    add_module(m->filename);
    
    if (remove) {
        /* when removing, we remove blindly all modules - the kernel
           will refure to unload the module still needed */
        remove_modules(modules_to_load, nb_modules_to_load);
    } else {
        /* find kernel based modules */
        get_loaded_modules();
        
        /* dump them */
        if (verbose) {
            const char *p;
            printf("Dependencies:");
            for(i = 0;i < nb_modules_to_load; i++) {
                printf(" %s", modules_to_load[i]->filename);
            }
            printf("\n");
            printf("Loaded modules:");
            p = loaded_modules;
            for(i = 0; i < nb_loaded_modules; i++) {
                printf(" %s", p);
                p += strlen(p) + 1;
            }
            printf("\n");
        }

        load_modules(modules_to_load, nb_modules_to_load);
    }

    return 0;
}
