/*
 * Load or remove a tiny module
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <elf.h>
#include <fcntl.h>
#include <linux/linkage.h>
#include <linux/module.h>
#include <errno.h>
#include <ctype.h>

#include <asm/unistd.h>

#include "tinsmod_load.h"

#define __NR_sys_init_module  __NR_init_module
#define __NR_sys_delete_module  __NR_delete_module

_syscall2(int, sys_init_module, const char *, name,
	  const struct module *, info)
_syscall1(int, sys_delete_module, const char *, name);

static struct module *load_fixed_module(const char *filename)
{
    int fd;
    long m_size, file_size;
    unsigned char *image;
    struct module *module;

    fd = open(filename, O_RDONLY);
    if (fd < 0)
	return NULL;
    file_size = lseek(fd, 0, SEEK_END);
    lseek(fd, 0, SEEK_SET);

    /* alloc module */
    image = malloc(file_size);
    if (!image)
	goto fail;
    if (read(fd, image, file_size) != file_size)
	goto fail;
    close(fd);

    m_size = ((struct module *) image)->size;

    /* realloc module to add .bss */
    if (m_size > file_size) {
	image = realloc(image, m_size);
	if (!image)
	    return NULL;
	memset(image + file_size, 0, m_size - file_size);
    }

    module = (struct module *) image;
    return module;
  fail:
    close(fd);
    return NULL;
}

static inline void *module_ptr(struct module *module, const void *ptr)
{
    return (char *)module + ((unsigned long)ptr - module->address);
}

static const char *get_modinfo_value(struct module *module, const char *key)
{
    const char *p, *v, *n, *ep;
    size_t klen = strlen(key);

    p = module_ptr(module, module->modinfo_start);
    ep = module_ptr(module, module->modinfo_end);
    while (p < ep) {
	v = strchr(p, '=');
	n = strchr(p, '\0');
	if (v) {
	    if (v - p == (int) klen && strncmp(p, key, klen) == 0)
		return v + 1;
	} else {
	    if (n - p == (int) klen && strcmp(p, key) == 0)
		return n;
	}
	p = n + 1;
    }

    return NULL;
}

/* I don't understand why parm_desc_xxx was choosen instead of
   desc_parm_xxx */
static int is_param(const char *p, const char *v)
{
    return (v && (v - p) >= 5 && memcmp(p, "parm_", 5) == 0 &&
            !((v - p)  >= 10 && memcmp(p, "parm_desc_", 10) == 0));
}

static unsigned long get_modinfo_data(struct module *module, const char *ep)
{
    const char *p, *n, *v;
    const unsigned long *pdata;

    pdata = module_ptr(module, module->modinfo_data);
    p = module_ptr(module, module->modinfo_start);
    for(;;) {
        n = strchr(p, '\0');
        v = strchr(p, '=');
        if (is_param(p, v)) {
            if (ep == (v + 1))
                break;
            pdata++;
        }
        p = n + 1;
    }
    return *pdata;
}

static int process_module_arguments(struct module **pmodule, 
                                    int argc, char **argv)
{
    struct module *module;
    int added_size;

    module = *pmodule;
    while (argc > 0) {
	const char *p, *q;
	char *loc, *key;
	int min, max, n;

	p = *argv;
	if ((q = strchr(p, '=')) == NULL)
            goto next;

	key = alloca(q - p + 6);
	memcpy(key, "parm_", 5);
	memcpy(key + 5, p, q - p);
	key[q - p + 5] = 0;

	p = get_modinfo_value(module, key);
	key += 5;
	if (p == NULL) {
	    error("invalid parameter %s", key);
	    return 0;
	}

        loc = module_ptr(module, (void *)get_modinfo_data(module, p));
        
	if (isdigit(*p)) {
	    min = strtoul(p, (char **)&p, 10);
	    if (*p == '-')
		max = strtoul(p + 1, (char **)&p, 10);
	    else
		max = min;
	} else
	    min = max = 1;

	n = (*++q != '\0');

	while (1) {
	    if ((*p == 's') || (*p == 'c')) {
		char *str;

		/* Do C quoting if we begin with a ", else slurp the lot.  */
		if (*q == '"') {
		    char *r;

		    str = alloca(strlen(q));
		    for (r = str, q++; *q != '"'; ++q, ++r) {
			if (*q == '\0') {
			    error
				("improperly terminated string argument for %s",
				 key);
			    return 0;
			} else if (*q == '\\')
			    switch (*++q) {
			    case 'a':
				*r = '\a';
				break;
			    case 'b':
				*r = '\b';
				break;
			    case 'e':
				*r = '\033';
				break;
			    case 'f':
				*r = '\f';
				break;
			    case 'n':
				*r = '\n';
				break;
			    case 'r':
				*r = '\r';
				break;
			    case 't':
				*r = '\t';
				break;

			    case '0':
			    case '1':
			    case '2':
			    case '3':
			    case '4':
			    case '5':
			    case '6':
			    case '7':
				{
				    int c = *q - '0';
				    if (q[1] >= '0' && q[1] <= '7') {
					c = (c * 8) + *++q - '0';
					if (q[1] >= '0' && q[1] <= '7')
					    c = (c * 8) + *++q - '0';
				    }
				    *r = c;
				}
				break;

			    default:
				*r = *q;
				break;
			} else
			    *r = *q;
		    }
		    *r = '\0';
		    ++q;
		} else {
		    char *r;

		    /* In this case, the string is not quoted. We will break
		       it using the coma (like for ints). If the user wants to
		       include comas in a string, he just has to quote it */

		    /* Search the next coma */
		    r = strchr(q, ',');

		    /* Found ? */
		    if (r != (char *) NULL) {
			/* Recopy the current field */
			str = alloca(r - q + 1);
			memcpy(str, q, r - q);

			/* I don't know if it is useful, as the previous case
			   doesn't null terminate the string ??? */
			str[r - q] = '\0';

			/* Keep next fields */
			q = r;
		    } else {
			/* last string */
			str = (char *)q;
			q = "";
		    }
		}

		if (*p == 's') {
                    int new_size;

		    /* Normal string : realloc the module to get some
                       space to put the string */
                    added_size = strlen(str) + 1;
                    loc -= (unsigned long)module;
                    new_size = module->size + added_size;
                    module = realloc(module, new_size);
                    if (!module) {
                        error("Not enough memory while allocating string");
                        return 0;
                    }
                    *pmodule = module;
                    loc += (unsigned long)module;

                    /* copy the string and patch the pointer to point
                       to it */
                    strcpy((char *)module + module->size, str);
                    *(char **)loc = (char *)module->address + module->size;
                    module->size = new_size;
		    loc += sizeof(char *);
		} else {
		    /* Array of chars (in fact, matrix !) */
		    long charssize;	/* size of each member */

		    /* Get the size of each member */
		    /* Probably we should do that outside the loop ? */
		    if (!isdigit(*(p + 1))) {
			error
			    ("parameter type 'c' for %s must be followed by"
			     " the maximum size", key);
			return 0;
		    }
		    charssize = strtoul(p + 1, (char **) NULL, 10);

		    /* Check length */
		    if (strlen(str) >= (size_t) charssize) {
			error("string too long for %s (max %ld)",
			      key, charssize - 1);
			return 0;
		    }

		    /* Copy to location */
		    strcpy((char *) loc, str);
		    loc += charssize;
		}
	    } else {
		long v = strtoul(q, (char **)&q, 0);
		switch (*p) {
		case 'b':
		    *loc++ = v;
		    break;
		case 'h':
		    *(short *) loc = v;
		    loc += sizeof(short);
		    break;
		case 'i':
		    *(int *) loc = v;
		    loc += sizeof(int);
		    break;
		case 'l':
		    *(long *) loc = v;
		    loc += sizeof(long);
		    break;

		default:
		    error("unknown parameter type '%c' for %s", *p, key);
		    return 0;
		}
	    }

	  retry_end_of_value:
	    switch (*q) {
	    case '\0':
		goto end_of_arg;

	    case ' ':
	    case '\t':
	    case '\n':
	    case '\r':
		++q;
		goto retry_end_of_value;

	    case ',':
		if (++n > max) {
		    error("too many values for %s (max %d)", key, max);
		    return 0;
		}
		++q;
		break;

	    default:
		error("invalid argument syntax for %s", key);
		return 0;
	    }
	}

      end_of_arg:
	if (n < min) {
	    error("too few values for %s (min %d)", key, min);
	    return 0;
	}
    next:
	argc--, argv++;
    }

    return 1;
}

int tiny_insmod(const char *filename, int argc, char **argv, int do_noload)
{
    char mod_name[128];
    struct module *module;
    
    /* load the module */
    module = load_fixed_module(filename);
    if (!module) {
        error("could not load '%s'", filename);
        return 1;
    }
    
    /* handle command line arguments */
    if (process_module_arguments(&module, argc, argv) == 0)
        return 1;
    strcpy(mod_name, module_ptr(module, module->name));
    
    if (!do_noload) {
        /* now we can load the module ! */
        if (sys_init_module(mod_name, module) < 0) {
            if (errno == EEXIST) {
                error("module '%s' is already loaded", mod_name);
            } else {
                /* XXX: maybe init_module should delete the module if
                   error, but it is more complicated to add */
                sys_delete_module(mod_name);
                perror(mod_name);
            }
            return 1;
        }
    }
    return 0;
}
