/* lcd.c
 *
 * $Id: lcd.c,v 1.9 2006-10-24 13:42:02 jma Exp $
 *
 * LCD driver for HD44780 compatible displays connected to the parallel port,
 * because real men use device files.
 *
 * Copyright (C) by Michael McLellan (mikey@mclellan.org.nz)
 * Released under the terms of the GNU GPL, see file COPYING for more details.
 */

#define KBUILD_MODNAME lcd

#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#ifndef KERNEL_VERSION
#define KERNEL_VERSION(a,b,c) (((a) << 16) + ((b) << 8) + (c))
#endif
#include <asm/uaccess.h>

#include <linux/proc_fs.h>
#include <linux/ioport.h>
#include <linux/module.h>
#include <linux/delay.h>
#include <linux/fs.h>
#include <asm/io.h>

#include <linux/hd44780.h>
#if defined(TRANSLATE_CHARCODES)
#include "charmap.h"
#endif

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
#include <linux/cdev.h>
#endif

#if defined(GPIO_IO)
#include <linux/gpio.h>

/* XXX: those IO are hardcoded for N12 */
static int bck_gpio;
#endif

#if defined (MODULE_LICENSE)
/* does MODULE_LICENSE exist in older kernel trees? */
MODULE_LICENSE("GPL");
#endif

#ifdef CGRAM_DEFAULT
#include "cgram/default.h"
#endif
#ifdef CGRAM_SWEDISH
#include "cgram/swedish.h"
#endif

#define MAX_DISP_ROWS	4		// The HD44780 supports up to 4 rows
#define MAX_DISP_COLS	40		// The HD44780 supports up to 40 columns
#define MAX_DISP_WRAP	40		// The HD44780 supports up to 40 columns
/* input_state states */
enum {
    LCD_STATE_NORMAL = 0,
    LCD_STATE_ESC,            // Escape sequence start
    LCD_STATE_DCA_Y,   	      // Direct cursor access, the next input will be the row
    LCD_STATE_DCA_X,          // Direct cursor access, the next input will be the column
    LCD_STATE_CGRAM_SELECT,   // Selecting which slot to enter new character
    LCD_STATE_CGRAM_ENTRY,    // Adding a new character to the CGRAM
    LCD_STATE_CGRAM_GENERATE, // Waiting fot the 8 bytes which define a character
    LCD_STATE_CHAR_MAP_OLD,   // Waiting for the original char to map to another
    LCD_STATE_CHAR_MAP_NEW,   // Waiting for the new char to replace the old one with
    LCD_STATE_ESC_,	      // Waiting for the [ in escape sequence
    LCD_STATE_ICON_ON,
    LCD_STATE_ICON_OFF,
    LCD_STATE_BCKLIGHT,
};

/* globals */
#if defined (MEM_MAPPED_IO)
static volatile unsigned char *io_mapped;
static int io = DFLT_BASE;
#endif
static int disp_rows = DFLT_DISP_ROWS;
static int disp_cols = DFLT_DISP_COLS;
static int disp_wrap = DFLT_DISP_WRAP;
//static int disp_wrap = DFLT_DISP_WRAP;
static unsigned char state[MAX_DISP_ROWS][MAX_DISP_COLS]; // The current state of the display
static int row = 0, column = 0; // Current virtual cursor position
static int wrap = 1; // Linewrap on/off
#ifdef __mips__
static char backlight_max = 2;
static char backlight = 2; // Backlight on-off
#else
static char backlight_max = 1;
static char backlight = 1; // Backlight on-off
#endif

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
static struct cdev *my_cdev; // character device
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
MODULE_DESCRIPTION("LCD driver for HD44780 compatable controllers");
MODULE_AUTHOR("Michael McLellan <mikey@mclellan.org.nz>");
#if defined (MEM_MAPPED_IO)
MODULE_PARM(io, "i");
#endif
MODULE_PARM(disp_rows, "i");
MODULE_PARM(disp_cols, "i");
MODULE_PARM(disp_wrap, "i");
#if defined (MEM_MAPPED_IO)
MODULE_PARM_DESC(io, "I/O base address (default: DEFAULT_BASE)");
#endif
MODULE_PARM_DESC(disp_rows,
                 "LCD rows (default: DFLT_DISP_ROWS, max: MAX_DISP_ROWS)");
MODULE_PARM_DESC(disp_cols,
                 "LCD columns (default: DFLT_DISP_COLS, max: MAX_DISP_COLS)");
MODULE_PARM_DESC(disp_wrap,
                 "LCD RAM wrap (default: DFLT_DISP_WRAP, max: MAX_DISP_WRAP)");
#endif

#if defined(USE_BACKLIGHT_TIMER)
static struct timer_list bck_timer;
static int bck_is_on;

static void bck_timer_func (unsigned long data)
{
    bck_is_on ^= 1;
#if defined (MEM_MAPPED_IO)
    /* XXX: TODO */
#elif defined(GPIO_IO)
    GPIO_Set(bck_gpio, bck_is_on);
#endif
    bck_timer.expires = jiffies + 1;
    add_timer(&bck_timer);
}
#endif

#if defined (MEM_MAPPED_IO)
/* Low level device access */
static inline void write_out (unsigned char data, int port)
{
#ifdef __powerpc__
    *(io_mapped + port) = data;
    __asm__ __volatile__ ("eieio");
#else
    outb(data, (unsigned long)io_mapped + port);
#endif
}

static inline unsigned char read_in (int port)
{
#ifdef __powerpc__
    __asm__ __volatile__ ("eieio");
    return *(io_mapped + port);
#else
    return inb((unsigned long)io_mapped + port);
#endif
}

static void set_backlight (void)
{
}

#elif defined(GPIO_IO)
static inline void write_out (unsigned char byte, int rs)
{
    bbspi_send_byte(byte, rs);
}

static void set_backlight (void)
{
    del_timer(&bck_timer);
    switch (backlight) {
    case 0:
        GPIO_Set(bck_gpio, 0);
        break;
    case 1:
#if defined(USE_BACKLIGHT_TIMER)
        bck_is_on = 1;
        bck_timer.expires = jiffies + 1;
        add_timer(&bck_timer);
        break;
#endif
        /* Fallback */
    case 2:
    default:
        GPIO_Set(bck_gpio, 1);
        break;
    }
}
#endif /* defined(GPIO_IO) */

static inline void wait_ready (void)
{
#if 0
    for (;;) {
        if ((read_in(0) & 0x80) == 0)
            break;
        udelay(10);
    }
#endif
}

/* Send an instruction to the display, e.g. move cursor */
static void writeCommand (unsigned char command)
{
    wait_ready();
    write_out(command, 0);
    /* Minimum time to wait after most commands is 39ns except Clear Display
     * and Return Home which is 1.53ms, we never use Return Home.
     */
    if (command <= 0x03 || 1)
        mdelay(2);
    else
        udelay(50);
}

static void writeData (unsigned char data)
{
    write_out(data, 1);
}

/* Send character data to the display, i.e. writing to DDRAM */
static void display_screen (unsigned char **buf, int cols, int rows)
{
    int i, j;

    writeCommand(0x02); // Cursor home
    for (i = 0; i < disp_rows; i++) {
        for (j = 0; j < disp_cols; j++) {
            wait_ready();
            writeData(state[i][j]);
            udelay(55);
        }
        for (; j < disp_wrap; j++) {
            wait_ready();
            writeData(0);
            udelay(55);
        }
    }
}

static void scroll_up (void)
{
    /* scroll up */
    memcpy(state, &state[1][0], disp_cols);
    memset(&state[1][0], 0x20, MAX_DISP_COLS);
}

static int display_char (unsigned char data)
{
    int ret;

    ret = 0;
    if (column < disp_cols) {
        if (state[row][column] != data) {
            state[row][column] = data;
            ret++;
        }
        column++;
    } else if (wrap) {
        if (row < disp_rows - 1) {
            row++;
        } else {
            scroll_up();
        }
        column = 0;
        if (state[row][column] != data) {
            state[row][column] = data;
            ret++;
        }
        column++;
    }

    return ret;
}

/* Write an entire (5x8) character to the CGRAM,
 * takes the CGRAM index, and a char[ 8 ] binary bitmap.
 */
static void writeCGRAM( int index, unsigned char pixels[] )
{
    int i;

    /* Move address pointer to index in CGRAM */
    writeCommand(0x40 + ( 8 * index ));
	for( i = 0; i < 8; i++ )
	{
		/* Write a line of a character. The E cycle time for
		 * a write should be at least 500 ns, and the minimum
		 * E pulse width should be 230 ns.
		 */
            //		writeData(E1_ON | RS_DATA | RW_WRITE | backlight);
            //		udelay( 250 );
            wait_ready();
		writeData( ( unsigned char ) pixels[ i ]);
                //		writeData( E1_OFF | RS_DATA | RW_WRITE | backlight);

		udelay( 250 );
		/* Time to wait after write to RAM is 43 ns */
		udelay( 45 );
	}
	row = column = -1;
}

#ifdef USE_ST7036
static int icons[ICON_MAX][2] = {
    { 0x03, 0x10, }, /* TV            */
    { 0x03, 0x08, }, /* Music         */
    { 0x07, 0x04, }, /* DTV           */
    { 0x07, 0x02, }, /* Tape          */
    { 0x07, 0x01, }, /* ATV           */
    { 0x0C, 0x10, }, /* DVD           */
    { 0x0C, 0x08, }, /* VCD           */
    { 0x0C, 0x01, }, /* MP3           */
    { 0x0C, 0x04, }, /* Pause         */
    { 0x0C, 0x02, }, /* Record        */
    { 0x0D, 0x02, }, /* Play          */
    { 0x0D, 0x01, }, /* Stop          */
    { 0x0E, 0x10, }, /* Start         */
    { 0x0E, 0x02, }, /* End           */
    { 0x0E, 0x08, }, /* Fast backward */
    { 0x0E, 0x04, }, /* Fast forward  */
    { 0x03, 0x01, }, /* Mute          */
    { 0x03, 0x02, }, /* TTX           */
    { 0x07, 0x08, }, /* DolbyR        */
    { 0x07, 0x10, }, /* DolbyL        */
    { 0x03, 0x04, }, /* Dollar        */
    { 0x0E, 0x01, }, /* Clock         */
};

static unsigned char icon_buf[0x10];

static void draw_icon (int icon_idx, int srt)
{
    int adr;

    if (icon_idx < 0 || icon_idx >= ICON_MAX)
        return;
    adr = icons[icon_idx][0];
    switch (srt) {
    case 0:
        /* Reset icon */
        icon_buf[adr] &= ~icons[icon_idx][1];
        break;
    case 1:
        /* Set icon */
        icon_buf[adr] |= icons[icon_idx][1];
        break;
    case 2:
    default:
        /* Toggle icon */
        icon_buf[adr] ^= icons[icon_idx][1];
        break;
    }
    writeCommand(0x31); // Function set
    writeCommand(0x40 | adr); // Set icon address
    udelay(1000);
    writeData(icon_buf[adr]);
    udelay(1000);
    writeCommand(0x30); // Function set
}
#else
static void draw_icon (int icon_idx, int srt)
{
}
#endif

static void initDisplay (void)
{
    /* initialize state array */
    memset(state, ' ', sizeof(state));
#ifdef USE_ST7036
    memset(icon_buf, 0, sizeof(icon_buf));
#endif
    /* initialize controller 1 */
    writeCommand(0x30);
    writeCommand(0x30);
    writeCommand(0x30);
#ifdef USE_ST7036
    writeCommand(0x08); // Display off
    writeCommand(0x31); // Function set
    writeCommand(0x1C); // Set bias
    writeCommand(0x7F); // Set contrast
    writeCommand(0x5F); // Power/icon/contrast
    writeCommand(0x6A); // Follower control
    writeCommand(0x30); // Function set
#else
    writeCommand(0x3C); // Function set
    writeCommand(0x08); // Display off
#endif
    writeCommand(0x01); // Clear display
    writeCommand(0x06); // Entry mode set
    writeCommand(0x02); // Cursor home
    writeCommand(0x0C); // Display on

    draw_icon(ICON_DTV, 1);

    /* Set the CGRAM to default values */
#if 0
    writeCGRAM(0, cg0);
    writeCGRAM(1, cg1);
    writeCGRAM(2, cg2);
    writeCGRAM(3, cg3);
    writeCGRAM(4, cg4);
    writeCGRAM(5, cg5);
    writeCGRAM(6, cg6);
    writeCGRAM(7, cg7);
    init_charmap();
#endif
}

static inline unsigned char char_remap (unsigned char c)
{
#if defined(TRANSLATE_CHARCODES)
    return charmap[c];
#else
    return c;
#endif
}


static int handleInput (unsigned char input)
{
    static int cgram_index = 0;
    static int cgram_row_count;
    static unsigned char cgram_pixels[8];
#if defined(TRANSLATE_CHARCODES)
    static unsigned char char_map_old;
#endif
    static int input_state = LCD_STATE_NORMAL; // the current state of the input handler
    int i;
    int temp, ret;

    ret = 0;
    switch (input_state) {
    case LCD_STATE_NORMAL:
        switch (input) {
        case 0x08: // Backspace
            if (column > 0) {
                column--;
                ret = display_char(' ');
                column--;
            }
            break;
        case 0x09: // Tabstop
            column = ((( column + 1) / TABSTOP) * TABSTOP) + TABSTOP - 1;
            break;
        case 0x0a: // Newline
            if (row < disp_rows - 1) {
                row++;
            } else {
                scroll_up();
            }
            /* Since many have trouble grasping the \r\n concept... */
            column = 0;
            break;
        case 0x0d: 	// Carrage return
            column = 0;
            break;
        case 0x1b: 	// esc ie. start of escape sequence
            input_state = LCD_STATE_ESC_;
            break;
        default:
            /* The character is looked up in the */
            ret = display_char(char_remap(input));
            break;
        }
        break;
    case LCD_STATE_ESC_:
        input_state = LCD_STATE_ESC;
        break;
    case LCD_STATE_ESC:
        if( input <= '7' && input >= '0' ) {
            /* Chararacter from CGRAM */
            ret = display_char(input - 0x30);
        } else {
            switch ( input ) {
            case 'A': // Cursor up
                if (row > 0)
                    row--;
                break;
            case 'B': // Cursor down
                if (row < disp_rows - 1)
                    row++;
                break;
            case 'C': // Cursor Right
                if (column < disp_cols - 1)
                    column++;
                break;
            case 'D': // Cursor Left
                if (column > 0)
                    column--;
                break;
            case 'H': // Cursor home
                row = 0;
                column = 0;
                break;
            case 'J': // Clear screen, cursor doesn't move
                memset(state, ' ', sizeof(state));
                writeCommand(0x01);
                break;
            case 'K': // Erase to end of line, cursor doesn't move
                temp = column;
                for (i = column; i < disp_cols; i++)
                    ret += display_char(' ');
                column = temp;
                break;
#if defined(TRANSLATE_CHARCODES)
            case 'M': // Charater mapping
                input_state = LCD_STATE_CHAR_MAP_OLD;
                break;
#endif
            case 'Y': // Direct cursor access
                input_state = LCD_STATE_DCA_Y;
                break;
            case 'R': // CGRAM select
                input_state = LCD_STATE_CGRAM_SELECT;
                break;
            case 'V': // Linewrap on
                wrap = 1;
                break;
            case 'W': // Linewrap off
                wrap = 0;
                break;
            case 'b': // Toggle backlight
                input_state = LCD_STATE_BCKLIGHT;
                break;
            case 'i':
                input_state = LCD_STATE_ICON_OFF;
                break;
            case 'I':
                input_state = LCD_STATE_ICON_ON;
                break;
            default:
                printk("LCD: unrecognized escape sequence: %#x ('%c')\n",
                       input, input);
            }
        }
        if (input_state != LCD_STATE_DCA_Y &&
            input_state != LCD_STATE_CGRAM_SELECT &&
            input_state != LCD_STATE_CHAR_MAP_OLD &&
            input_state != LCD_STATE_ICON_OFF &&
            input_state != LCD_STATE_ICON_ON  &&
            input_state != LCD_STATE_BCKLIGHT) {
            input_state = LCD_STATE_NORMAL;
        }
        break;
    case LCD_STATE_DCA_Y:
        if ( input - 0x1f < disp_rows )
            row = input - 0x1f;
        else {
            printk( "LCD: tried to set cursor to off screen location\n" );
            row = disp_rows - 1;
        }
        input_state = LCD_STATE_DCA_X;
        break;
    case LCD_STATE_DCA_X:
        if ( input - 0x1f < disp_cols ) {
            column = input - 0x1f;
        } else {
            printk( "LCD: tried to set cursor to off screen location\n" );
            column = disp_cols - 1;
        }
        input_state = LCD_STATE_NORMAL;
        break;
    case LCD_STATE_CGRAM_SELECT:
        if( input > '7' || input < '0' ) {
            printk( "LCD: Bad CGRAM index %c\n", input );
            input_state = LCD_STATE_NORMAL;
        } else {
            cgram_index = input - 0x30;
            cgram_row_count = 0;
            input_state = LCD_STATE_CGRAM_GENERATE;
        }
        break;
    case LCD_STATE_CGRAM_GENERATE:
        cgram_pixels[cgram_row_count++ ] = input;
        if( cgram_row_count == 8) {
            writeCGRAM(cgram_index, cgram_pixels);
            input_state = LCD_STATE_NORMAL;
        }
        break;
#if defined(TRANSLATE_CHARCODES)
    case LCD_STATE_CHAR_MAP_OLD:
        char_map_old = input;
        input_state = LCD_STATE_CHAR_MAP_NEW;
        break;
    case LCD_STATE_CHAR_MAP_NEW:
        charmap[char_map_old] = input;
        input_state = LCD_STATE_NORMAL;
        break;
#endif
    case LCD_STATE_ICON_OFF:
        draw_icon(input, 0);
        input_state = LCD_STATE_NORMAL;
        break;
    case LCD_STATE_ICON_ON:
        draw_icon(input, 1);
        input_state = LCD_STATE_NORMAL;
        break;
    case LCD_STATE_BCKLIGHT:
        if (input < '0') {
            backlight = 0;
        } else if (input > '9') {
            backlight = backlight_max;
        } else {
            backlight = input - '0';
            if (backlight > backlight_max)
                backlight = backlight_max;
        }
        set_backlight();
        input_state = LCD_STATE_NORMAL;
        break;
    default:
        input_state = LCD_STATE_NORMAL;
        break;
    }

    return ret;
}

/* Handle device open */
int lcd_open (struct inode *minode, struct file *mfile)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    if (MOD_IN_USE)
        return -EBUSY;
    MOD_INC_USE_COUNT;
#else
// FIXME : 2.6.x inc use count
#endif

    return 0;
}

/* Handle device file close */
int lcd_release (struct inode *minode, struct file *mfile)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    MOD_DEC_USE_COUNT;
#else
// FIXME : 2.6.x dec use count
#endif

    return 0;
}

/* Handle write to device file */
ssize_t lcd_write_byte (struct file *inode, const char *gdata,
                        size_t length, loff_t *off_what)
{
    int i;
    int b;
    int ret, n;

    n = 0;
    for (i = 0; i < length; i++) {
	if (get_user(b, gdata + i)) {
	    return -EFAULT;
	}
        ret = handleInput(b);
        if (ret < 0)
            break;
        n += ret;
    }
    if (n > 0)
        display_screen((unsigned char **)state, MAX_DISP_COLS, MAX_DISP_ROWS);
    
    return length;
}

/* Handle read from device file */
ssize_t lcd_read_byte (struct file *inode, char *udata,
                       size_t length, loff_t *loff_what)
{
    return 0;
}

/* Handle read from proc file */
int readProcFile (char *buffer, char **start, off_t offset,
                  int size, int *eof, void *data)

{
    char *temp = buffer;
    int i, j;
        
    /* Print module configuration */
    temp += sprintf(temp,
#if defined (MEM_MAPPED_IO)
                    "I/O base:        %#x\n"
#endif
                    "Display rows:    %d\n"
                    "Display columns: %d\n"
                    "Linewrap:        %s\n"
                    "Pos : %d - %d\n",
#if defined (MEM_MAPPED_IO)
                    io,
#endif
                    disp_rows, disp_cols,
                    wrap?"On":"Off", column, row);
    /* Print display state */
    temp += sprintf(temp, "+");
    for (i = 0; i < disp_cols; i++)
        temp += sprintf(temp, "-");
    temp += sprintf(temp, "+\n");
    
    for (i = 0; i < disp_rows; i++) {
        temp += sprintf(temp, "|");
        for (j = 0; j < disp_cols ; j++)
            temp += sprintf(temp, "%c",
                            (state[i][j] < 10) ? '' : state[i][j]);
        temp += sprintf(temp, "|\n");
    }
    temp += sprintf(temp, "+");
    for (i = 0; i < disp_cols; i++)
        temp += sprintf(temp, "-");
    temp += sprintf(temp, "+\n");
    return temp - buffer;
}

/* Module cleanup */
void cleanup_module (void)
{
#if defined(USE_BACKLIGHT_TIMER)
    del_timer(&bck_timer);
#endif
    backlight = 0;
    set_backlight();
    writeCommand(0x08); // Display off
    writeCommand(0x01); // Clear display
#if defined (MEM_MAPPED_IO)
    release_region(io, 3);
#endif
    remove_proc_entry("lcd", 0);
#if defined(GPIO_IO)
    bbspi_reset();
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    if (unregister_chrdev(LCD_MAJOR, "LCD panel"))
        printk("LCD: module cleanup failed\n");
    else
        printk("LCD: module cleanup OK\n");
#else
    dev_t dev = MKDEV(LCD_MAJOR,0);
    printk("LCD: deleting char device %d\n", dev);
    // these functions return void so no error checking possible
    unregister_chrdev_region(dev, 1);
    cdev_del(my_cdev);
#endif
}

/* Module initilisation */
int init_module (void)
{

#if  LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    static struct file_operations lcd_fops =
        { NULL, NULL, lcd_read_byte, lcd_write_byte,
          NULL, NULL, NULL, NULL, lcd_open, NULL, lcd_release, };
#else 
    static struct file_operations lcd_fops =
        { NULL, NULL, lcd_read_byte, NULL, lcd_write_byte,
          NULL, NULL, NULL, NULL, NULL, lcd_open, NULL, lcd_release, };
    dev_t dev;
#endif
    int retcode;
#if defined (MEM_MAPPED_IO)
    int io_len;
#endif
    
    retcode = -EACCES;
#if defined (MEM_MAPPED_IO)
#ifdef __powerpc__
    io_len = 4096;
#else
    io_len = 16;
#endif
    if (check_region(io, io_len)) {
        printk("LCD: check_region failed, io = %#x\n", io);
        return retcode;
    }
#if 0
    if (request_region(io, io_len, "LCD panel")) {
        printk("LCD: request_region failed, io = %#x\n", io);
        return retcode;
    }
#endif
#ifdef __powerpc__
    io_mapped = ioremap(io, io_len);
#else
    io_mapped = (unsigned char *)io;
#endif
#endif
#if defined(GPIO_IO)
    bbspi_init(3, 2, 4, 5);
    bck_gpio = 22;
#endif

    /* Make sure user didn't pass silly numbers, MAX_DISP_???? are just
     * arbitary numbers and can be increased if need be.
     */
    if (disp_rows > MAX_DISP_ROWS)
        disp_rows = MAX_DISP_ROWS;
    if (disp_cols > MAX_DISP_COLS)
        disp_cols = MAX_DISP_COLS;
    if (disp_wrap > MAX_DISP_COLS)
        disp_wrap = MAX_DISP_COLS;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    if (register_chrdev(LCD_MAJOR, "LCD panel", &lcd_fops)) {
        printk("LCD: Error! Unable to bind to major %d\n", LCD_MAJOR );
        goto error;
    }
#else 
    // device number
    dev = MKDEV(LCD_MAJOR, 0);
    if (register_chrdev_region(dev, 1, "LCD panel")) {
        printk("LCD: char device register failed\n"); 
        goto error;
    }
    printk ("LCD: char device registered\n"); 
    // fill cdev structure
    my_cdev = cdev_alloc();
    my_cdev->owner = THIS_MODULE;
    my_cdev->ops = &lcd_fops;
    kobject_set_name(&my_cdev->kobj, "LCD panel");
    
    // register it
    if (cdev_add(my_cdev, dev, 1)) { 
        printk("LCD char device add failed\n");
        goto error;
    } 
    printk("LCD: char device added\n"); 
#endif // LINUX_VERSION_CODE

    initDisplay();
#if defined(USE_BACKLIGHT_TIMER)
    init_timer(&bck_timer);
    bck_timer.function = bck_timer_func;
#endif
    backlight = backlight_max;
    set_backlight();
    retcode = -ENOMEM;
    if (!create_proc_read_entry("lcd", 0, 0, readProcFile, NULL)) {
        printk(KERN_ERR "LCD: Error! Can't create /proc/lcd\n");
        goto error;
    }
    printk("LCD: init OK, "
#if defined (MEM_MAPPED_IO)
           "iobase: %#x (%p), "
#endif
           "rows: %d, columns: %d\n",
#if defined (MEM_MAPPED_IO)
           io, io_mapped,
#endif
           disp_rows, disp_cols);

    return 0;

 error:
#if defined (MEM_MAPPED_IO)
#ifdef __powerpc__
    iounmap((void *)io_mapped);
#endif
    release_region(io, io_len);
#endif
#if defined(GPIO_IO)
    bbspi_reset();
#endif

    return retcode;
}
