#define __NO_STRING_INLINES
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <errno.h>

#include <zlib.h>

#include "minixfs.h"
#include "cminix.h"

#define MINIX_BLOCK_SIZE   1024
#define MINIX_HEADER_SIZE  7

/* return the compressed size. If it is cluster_size, outbuf is a copy
   of inbuf */
static int minix_compress_cluster(unsigned char *outbuf,
                                  unsigned char *inbuf,
                                  int cluster_size)
{
    z_stream strm;
    int ret,out_len;
    unsigned long crc;

    strm.zalloc = Z_NULL;
    strm.zfree = Z_NULL;

    /* best compression, small window, no zlib header */
    ret = deflateInit2(&strm, Z_DEFAULT_COMPRESSION,
                       Z_DEFLATED, -MINIX_CLUSTER_BITS, 
                       9, Z_DEFAULT_STRATEGY);
    
    if (ret != 0) goto uncompressed;

    /* header */
    crc = adler32(0, inbuf, cluster_size);
    /* magic */
    outbuf[0] = (MINIX_COMPR_MAGIC >> 8) & 0xff;
    outbuf[1] = (MINIX_COMPR_MAGIC) & 0xff;
    /* method */
    outbuf[2] = 0;
    /* crc */
    outbuf[3] = crc & 0xff;
    outbuf[4] = (crc >> 8) & 0xff;
    outbuf[5] = (crc >> 16) & 0xff;
    outbuf[6] = (crc >> 24) & 0xff;

    strm.avail_in = cluster_size;
    strm.next_in = inbuf;
    strm.avail_out = cluster_size - MINIX_HEADER_SIZE;
    strm.next_out = outbuf + MINIX_HEADER_SIZE;

    ret = deflate(&strm, Z_FINISH);
    if (ret != Z_STREAM_END && ret != Z_OK) {
        deflateEnd(&strm);
        return -1;
    }
    out_len = cluster_size - strm.avail_out;

    deflateEnd(&strm);

    if (ret != Z_STREAM_END || out_len >= cluster_size) {
        /* could not compress : copy the input */
    uncompressed:
        out_len = cluster_size;
        memcpy(outbuf, inbuf, out_len);
    }

    return out_len;
}

static unsigned long cpu_to_le32(unsigned long l)
{
#if (__BYTE_ORDER == __LITTLE_ENDIAN)
    return l;
#else
    return 
        ((l & 0xFF000000) >> 24) |
        ((l & 0x00FF0000) >> 8) |
        ((l & 0x0000FF00) << 8) |
        ((l & 0x000000FF) << 24);
#endif
}

/* Minix-compress file F of length FILESIZE to a created file FILEPATH.  */
/* return < 0 if error, == 0 if OK */
int copy_fd_file_minix_compress(int f, char *filepath, int filesize)
{
    int fd,cluster_index;
    int ret, len,cluster_size, pos, out_len,size, nb_offsets, tab_size;
    unsigned char inbuf[MINIX_CLUSTER_SIZE];
    unsigned char outbuf[MINIX_CLUSTER_SIZE];
    unsigned long *tab;
    int update = 1;

    fd = -1; /* fail safe */
    cluster_size = MINIX_CLUSTER_SIZE;
    size = filesize;

    if (update) {
        /* we unlink the file in case it already exists (bug of
           compressed minix possible) */
        unlink(filepath);

        fd = open(filepath, O_WRONLY | O_CREAT | O_TRUNC, 0666);
        if (fd < 0) {
            perror(filepath);
            return -1;
        }
    }
    
    /* we add one offset to mark the end of the compressed file */
    nb_offsets = ((filesize + cluster_size - 1) / cluster_size) + 1;
    tab_size = nb_offsets * sizeof(tab[0]);
    tab = malloc (tab_size);

    if (update) {
        /* skip the table size */
        lseek(fd, tab_size, SEEK_SET);
    }

    cluster_index = 0;
    pos = tab_size;
    while (size > 0) {
        len = size;
        if (len > cluster_size) len = cluster_size;
        tab[cluster_index] = cpu_to_le32(pos);
        memset(inbuf, 0, cluster_size);
        ret = read(f, inbuf, len);
        if (len != ret) goto error;
        
        out_len = minix_compress_cluster(outbuf, inbuf, cluster_size);
        if (out_len < 0)
            goto error;

        if (update) {
            ret = write(fd, outbuf, out_len);
            if (ret != out_len) goto error;
        }
        size -= len;
        pos += out_len;
        cluster_index++;
    }
    tab[cluster_index] = cpu_to_le32(pos);

    if (pos >= filesize) {
        /* if the compressed file is bigger than the original, we must
           not compress because the size will be incorrect */

        /* XXX: very big problem, should use a big cluster size */
        fprintf(stderr,"minix: compressed file bigger than original\n");
        goto error;
    }

    if (update) {
        /* write the offset table */
        lseek(fd, 0, SEEK_SET);
        write(fd, tab, tab_size);
    }

    free(tab);

    if (update) {
#if 0
        /* we expand the file so that is has its real size */
        ftruncate(fd, filesize);

        /* now we set the file as being compressed */
        tmp = MINIX_COMPRBLK_4KB; 
        if (ioctl(fd, MINIX_IOC_SETFLAGS, &tmp) < 0) {
            perror("MINIX_IOC_SETFLAGS");
        }
#endif
        close(fd);
    }
    
    return 0;
 error:
    perror(filepath);
    if (update) {
        close(fd);
    }
    return -1;
}
