 /*
 * SDPL read.c: routines able to parse a sdpl layer
 *
 *   Copyright (C) 2001 - 2007, Jocelyn MAYER <l_indien at magic dot fr>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 */
/*
 * Simple Disk Partition Layer
 *
 * Retrieve layer infos routines
 *
 */

/**SECTION: sysutils */
/**SECTION: sdpl : sysutils */
/**SECTION: editTool : sdpl */

#ifdef __KERNEL__
#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/unistd.h>
#include <linux/slab.h>
#include <linux/hdreg.h>
#include <linux/file.h>
#include <asm/uaccess.h>
#include <linux/blk.h>
/* We need adler32 from zlib */
#if CONFIG_UNIZLIB
#include <linux/zlib.h>
#else
#include "adler32.h"
#endif

#else /* __KERNEL__ */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#ifdef CONFIG_BLK_SDPL_64BITS
/* These are the GLIBC definition to use 64 bits file support */
#ifndef __USE_LARGEFILE64
#define __USE_LARGEFILE64 1
#endif
#ifndef __USE_FILE_OFFSET64
#define __USE_FILE_OFFSET64 1
#endif
#else /* CONFIG_BLK_SDPL_64BITS */
#ifndef __USE_LARGEFILE64
#define __USE_LARGEFILE64 0
#endif
#ifndef __USE_FILE_OFFSET64
#define __USE_FILE_OFFSET64 0
#endif
#endif /* CONFIG_BLK_SDPL_64BITS */
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <netinet/in.h>
#include <limits.h>
#include <linux/types.h>
/* We need adler32 from zlib */
#include <zlib.h>

#endif /* __KERNEL__ */

#include <linux/version.h>
#include <linux/sdpl.h>

#ifdef __KERNEL__

/* This function ISN'T exported for modules */
#if defined(MODULE) || (LINUX_VERSION_CODE < 0x2040A)
unsigned char *read_dev_sector(block_dev_t *bdev, unsigned long n,
                               Sector *p)
{
    int bloc = n * 512 / PAGE_CACHE_SIZE;
#if (LINUX_VERSION_CODE < 0x2040A)
    struct page *page;

    page = bread(bdev->device, bloc, PAGE_CACHE_SIZE)->b_page;
    p->v = page;
    return BUFFER(p, n);
#else
    struct address_space *mapping = bdev->bd_inode->i_mapping;
    struct page *page;

    page = read_cache_page(mapping, bloc,
                           (filler_t *)mapping->a_ops->readpage, NULL);
    if (!IS_ERR(page)) {
        wait_on_page(page);
        if (!Page_Uptodate(page))
            goto fail;
        if (PageError(page))
            goto fail;
        p->v = page;
        return BUFFER(p, n);
    fail:
        page_cache_release(page);
    }
    p->v = NULL;
    return NULL;
#endif
}
#endif /* MODULE */

#else /* __KERNEL__ */

/** Low level read/write routines
 *  This routine wants to have the same prototype
 *  than the one used in the linux kernel (fs/partition/check.c)
 */
unsigned char *read_dev_sector (block_dev_t *bdev, unsigned long n,
                                Sector *p)
{
    sdpl_lseek(bdev->fd, n * 512, SEEK_SET);
    if (sdpl_read(bdev->fd, p->sector, 512) < 512)
        return NULL;

    return p->sector;
}

#endif /* __KERNEL__ */

/** Retrieve layer's informations
 */
int get_SDPL_infos (struct SDPL_map *SDPL_map)
{
    SDPL_br_t *MBR = SDPL_map->mbr;
    union SDPL_infos *infos;
    u32 check;

#if CONFIG_SDPL_USE_MBR_COPY
    int tries = 0;

 retry:
#endif
    sdpl_debug("Get magic... (%p)\n", MBR);
    SDPL_map->magic = ntohs(MBR->magic);
    SDPL_map->infos = (void *)MBR->SDPL_infos;
    infos = SDPL_map->infos;
    if ((SDPL_map->magic & 0xFF00) != 0xE700) {
        sdpl_error("File don't contain any SDpl mapping (%08X) !\n",
                   SDPL_map->magic);
#if CONFIG_SDPL_USE_MBR_COPY
        goto do_retry;
#else
        return -1;
#endif
    }
    sdpl_debug("Check CRC\n");
    SDPL_map->CRC = ntohl(MBR->CRC);
    check = 1;
    check = (unsigned int)adler32(check, ((char *)MBR + 8), 512 - 8);
    if (SDPL_map->CRC != check) {
        sdpl_error("Bad CRC (%04X - %04X)\n", SDPL_map->CRC, check);
#if CONFIG_SDPL_USE_MBR_COPY
        goto do_retry;
#else
        return -1;
#endif
    }
    SDPL_map->info_size = ntohs((u16)MBR->info_size);
    SDPL_map->bloc_dec = 0;
    sdpl_debug("Get infos\n");
    switch (SDPL_map->info_size) {
    case 16:
        sdpl_debug("Get embbedded infos\n");
#if !CONFIG_BLK_SDPL_EMBEDDED
        EMBEDDED(SDPL_map) = 1;
        sdpl_debug("Get nparts\n");
        NPARTS(SDPL_map) = ntohs(infos->infos16.nparts);
#endif /* !CONFIG_BLK_SDPL_EMBEDDED */
        SDPL_map->bloc_size = ntohs(infos->infos16.bloc_size);
        SDPL_map->SDPL_size = ntohs(infos->infos16.SDPL_size);
        SDPL_map->pentries_offset = ntohs(infos->infos16.table_offset);
        SDPL_map->bitmap_offset = ntohs(infos->infos16.bitmap_offset);
        SDPL_map->data_offset = ntohs(infos->infos16.data_offset);
        if (SDPL_map->pentries_offset != 0x08 ||
            SDPL_map->bitmap_offset != 0x190) {
#if !CONFIG_BLK_SDPL_EMBEDDED
            EMBEDDED(SDPL_map) = 0;
#else /* !CONFIG_BLK_SDPL_EMBEDDED */
#if CONFIG_SDPL_USE_MBR_COPY
            goto do_retry;
#else
            return -1;
#endif
#endif /* !CONFIG_BLK_SDPL_EMBEDDED */
        }
        break;
#if !CONFIG_BLK_SDPL_EMBEDDED
    case 32:
        sdpl_debug("Get 32 bits infos\n");
        SDPL_map->bloc_size = ntohl(infos->infos32.bloc_size);
        sdpl_debug("Get nparts\n");
        SDPL_map->SDPL_size = ntohl(infos->infos32.SDPL_size);
        sdpl_debug("Get nparts\n");
        NPARTS(SDPL_map) = ntohl(infos->infos32.nparts);
        SDPL_map->pentries_offset = ntohl(infos->infos32.table_offset);
        SDPL_map->bitmap_offset = ntohl(infos->infos32.bitmap_offset);
        SDPL_map->data_offset = ntohl(infos->infos32.data_offset);
        break;
#if CONFIG_BLK_SDPL_64BITS
    case 64:
        sdpl_debug("Get 64 bits infos\n");
        SDPL_map->bloc_size = ntohll(infos->infos64.bloc_size);
        SDPL_map->SDPL_size = ntohll(infos->infos64.SDPL_size);
        sdpl_debug("Get nparts\n");
        NPARTS(SDPL_map) = ntohll(infos->infos64.nparts);
        SDPL_map->pentries_offset = ntohll(infos->infos64.table_offset);
        SDPL_map->bitmap_offset = ntohll(infos->infos64.bitmap_offset);
        SDPL_map->data_offset = ntohll(infos->infos64.data_offset);
        break;
#endif /* CONFIG_BLK_SDPL_64BITS */
#endif /* !CONFIG_BLK_SDPL_EMBEDDED */
    default:
        sdpl_error("Incorrect data size for SDpl (%d)\n", SDPL_map->info_size);
#if CONFIG_SDPL_USE_MBR_COPY
        goto do_retry;
#else
        return -1;
#endif
    }
    if ((SDPL_map->bloc_size & (SDPL_map->bloc_size - 1)) == 0) {
        std_type size = SDPL_map->bloc_size;
        SDPL_map->bloc_dec = 1;
        while (size != 2) {
            size /= 2;
            SDPL_map->bloc_dec++;
        }
    }
    {
#if !CONFIG_BLK_SDPL_EMBEDDED
        u64 lost;
        lost = (u64)SDPL_map->device_size * 2;
        lost -= (u64)SDPL_map->SDPL_size * SDPL_map->bloc_size;
        SDPL_map->res_size = lost;
        sdpl_debug("Reserved size is: %llu (" STD_ATTRIB " "
                   STD_ATTRIB ")\n", SDPL_map->res_size, SDPL_map->device_size,
                   SDPL_map->SDPL_size);
#else
        u32 lost;
        lost = (u32)SDPL_map->device_size * 2;
        lost -= (u32)SDPL_map->SDPL_size * SDPL_map->bloc_size;
        SDPL_map->res_size = lost;
        sdpl_debug("Reserved size is: %u (" STD_ATTRIB " "
                   STD_ATTRIB ")\n", SDPL_map->res_size, SDPL_map->device_size,
                   SDPL_map->SDPL_size);
#endif
    }

    return 0;

#if CONFIG_SDPL_USE_MBR_COPY
 do_retry:
    if (tries == 0) {
        sdpl_error("Changing MBR from %08X", (unsigned int)MBR);
        MBR = (SDPL_br_t *)((char *)MBR + 512);
        SDPL_map->mbr = MBR;
        tries++;
        sdpl_error(" to %08X\n", (unsigned int)MBR);
        goto retry;
    }
    return -1;
#endif
}

/* Get infos about partitions */
void get_parts_infos (struct SDPL_map *SDPL_map, Sector sect,
                      block_dev_t *bdev)
{
    /* Partitions infos */
    std_type pnext = 0, off_max = 1024;
    /* We have to read the right sector... */
#if !CONFIG_BLK_SDPL_EMBEDDED
    std_type cur_sect = 0, wanted_sect = 0;
#endif /* !CONFIG_BLK_SDPL_EMBEDDED */
    std_type offset;
    std_type minor, bloc, j;
    union part_entry *pentry;
    unsigned char *pmap;
    unsigned char *bitmap;
    unsigned char *buffer;

    /* Allocate all structures */
    /* In EDIT case, we wanna be able to realloc each part... */
    SDPL_map->pentries =
        sdpl_malloc(NPARTS(SDPL_map) * sizeof(union part_entry *));
    pnext = SDPL_map->pentries_offset;
    SDPL_map->pmagic = sdpl_malloc(NPARTS(SDPL_map) * sizeof(u16));
    SDPL_map->pflags = sdpl_malloc(NPARTS(SDPL_map) * sizeof(u16));
    SDPL_map->psize = sdpl_malloc(NPARTS(SDPL_map) * sizeof(std_type));
    SDPL_map->poffset = sdpl_malloc(NPARTS(SDPL_map)  * sizeof(std_type));
    SDPL_map->pnext = sdpl_malloc(NPARTS(SDPL_map)  * sizeof(std_type));
    SDPL_map->pbloc = sdpl_malloc(NPARTS(SDPL_map)  * sizeof(std_type *));
    SDPL_map->bflags = sdpl_malloc(SDPL_map->SDPL_size * sizeof(u16));
    memset(SDPL_map->bflags, EMPTY_FILL, SDPL_map->SDPL_size * sizeof(u16));

    minor = 0;
    buffer = (unsigned char *)SDPL_map->mbr;
    while (minor < NPARTS(SDPL_map)) {
        /* In an embedded layer, all is grouped in the first sector */
#if !CONFIG_BLK_SDPL_EMBEDDED
        off_max = max(off_max, (std_type)(pnext + 0x10));
        /* Don't forget to get the right sector */
        if ((wanted_sect = (pnext / 512)) != cur_sect) {
            sdpl_debug("get part: Sector " STD_ATTRIB
                       " to read... (%08X)\n", wanted_sect,
                       (unsigned int)pnext);
            put_dev_sector(sect);
            buffer = read_dev_sector(bdev, wanted_sect, &sect);
            cur_sect = wanted_sect;
        }
#endif /* !CONFIG_BLK_SDPL_EMBEDDED */
        SDPL_map->pentries[minor] = (void *)((pnext & 511) + buffer);

        pentry = SDPL_map->pentries[minor];
#if !CONFIG_BLK_SDPL_EMBEDDED
        switch(SDPL_map->info_size) {
        case 16:
#endif /* !CONFIG_BLK_SDPL_EMBEDDED */
            SDPL_map->pmagic[minor] = ntohs(pentry->pentry16.magic);
            SDPL_map->pflags[minor] = ntohs(pentry->pentry16.flags);
            SDPL_map->psize[minor] = ntohs(pentry->pentry16.size);
            SDPL_map->poffset[minor] = ntohs(pentry->pentry16.map_offset);
            pnext = ntohs(pentry->pentry16.next_part);
#ifndef __KERNEL__
            if (SDPL_map->poffset[minor] != 0x48 + minor * 0x50) {
#if !CONFIG_BLK_SDPL_EMBEDDED
                EMBEDDED(SDPL_map) = 0;
#else /* !CONFIG_BLK_SDPL_EMBEDDED */
                return;
#endif /* !CONFIG_BLK_SDPL_EMBEDDED */
            }
#endif /* __KERNEL__ */
#if !CONFIG_BLK_SDPL_EMBEDDED
            break;
        case 32:
            SDPL_map->pmagic[minor] = ntohs(pentry->pentry32.magic);
            SDPL_map->pflags[minor] = ntohs(pentry->pentry32.flags);
            SDPL_map->psize[minor] = ntohl(pentry->pentry32.size);
            SDPL_map->poffset[minor] = ntohl(pentry->pentry32.map_offset);
            pnext = ntohl(pentry->pentry32.next_part);
            break;
#if CONFIG_BLK_SDPL_64BITS
        case 64:
            SDPL_map->pmagic[minor] = ntohs(pentry->pentry64.magic);
            SDPL_map->pflags[minor] = ntohs(pentry->pentry64.flags);
            SDPL_map->psize[minor] = ntohll(pentry->pentry64.size);
            SDPL_map->poffset[minor] = ntohll(pentry->pentry64.map_offset);
            pnext = ntohll(pentry->pentry64.next_part);
            break;
#endif /* CONFIG_BLK_SDPL_64BITS */
        }
#endif /* !CONFIG_BLK_SDPL_EMBEDDED */
        SDPL_map->pnext[minor] = pnext;
        offset = SDPL_map->poffset[minor];
        SDPL_map->pbloc[minor] =
            sdpl_malloc(SDPL_map->psize[minor] * sizeof(std_type));
        for (j = 0; j < SDPL_map->psize[minor]; j++) {
            /* Still the same problem... */
#if !CONFIG_BLK_SDPL_EMBEDDED
            if ((wanted_sect = (offset / 512)) != cur_sect) {
                sdpl_debug("pmap: Sector " STD_ATTRIB " to read ("
                           STD_ATTRIB ")...\n", wanted_sect, offset);
                put_dev_sector(sect);
                buffer = read_dev_sector(bdev, wanted_sect, &sect);
                cur_sect = wanted_sect;
            }
#endif /* !CONFIG_BLK_SDPL_EMBEDDED */
            pmap = (buffer + (offset & 511));
#if !CONFIG_BLK_SDPL_EMBEDDED
            switch(SDPL_map->info_size) {
            case 16:
#endif /* !CONFIG_BLK_SDPL_EMBEDDED */
                SDPL_map->pbloc[minor][j] = ntohs(*((u16 *)pmap));
                offset += 2;
#if !CONFIG_BLK_SDPL_EMBEDDED
                break;
            case 32:
                SDPL_map->pbloc[minor][j] = ntohl(*((u32 *)pmap));
                offset +=4 ;
                break;
#if CONFIG_BLK_SDPL_64BITS
            case 64:
                SDPL_map->pbloc[minor][j] = ntohll(*((u64 *)pmap));
                offset += 8;
                break;
#endif /* CONFIG_BLK_SDPL_64BITS */
            }
#endif /* !CONFIG_BLK_SDPL_EMBEDDED */
            /* To avoid crashes in kernel space, we need to check... */
            if (SDPL_map->pbloc[minor][j] >= SDPL_map->SDPL_size) {
                /* Ignore this bloc */
                sdpl_error("Inconsistent bloc number: " STD_ATTRIB " - "
                           STD_ATTRIB "\n", SDPL_map->pbloc[minor][j],
                           SDPL_map->psize[minor]);
                j--;
                SDPL_map->psize[minor]--;
            } else {
                SDPL_map->bflags[SDPL_map->pbloc[minor][j]] =
                    SDPL_map->pflags[minor] & (SDPL_RO | SDPL_RW);
            }
        }
        off_max = max(off_max, offset);
        minor++;
    }
    /* Load blocs flags */
    offset = SDPL_map->bitmap_offset;
    buffer = (unsigned char *)SDPL_map->mbr;
#if !CONFIG_BLK_SDPL_EMBEDDED
    if ((wanted_sect = (offset / 512)) != cur_sect) {
        sdpl_debug("Pmap: Sector " STD_ATTRIB " to read ("
                   STD_ATTRIB ")...\n", wanted_sect, offset);
        put_dev_sector(sect);
        buffer = read_dev_sector(bdev, wanted_sect, &sect);
        cur_sect = wanted_sect;
    }
#endif /* !CONFIG_BLK_SDPL_EMBEDDED */
    j = 0;
    for (bloc = 0; bloc < SDPL_map->SDPL_size; bloc++) {
        bitmap = buffer + (offset & 511);
        /* If RO or empty, don't take care of bitmap */
        if (SDPL_map->bflags[bloc] != EMPTY_BLOC)
            SDPL_map->bflags[bloc] = (((*bitmap >> j) & 1) == 1 ?
                                      SDPL_RW : SDPL_RO);
        j++;
        if (j == 8) {
            offset++;
            j = 0;
#if !CONFIG_BLK_SDPL_EMBEDDED
            if ((wanted_sect = (offset / 512)) != cur_sect) {
                sdpl_debug("Pmap: Sector " STD_ATTRIB " to read ("
                           STD_ATTRIB ")...\n", wanted_sect, offset);
                put_dev_sector(sect);
                buffer = read_dev_sector(bdev, wanted_sect, &sect);
                cur_sect = wanted_sect;
            }
#endif /* !CONFIG_BLK_SDPL_EMBEDDED */
        }
    }
}

#if CONFIG_BLK_SDPL_DEBUG > 2 || !defined(__KERNEL__)
void dump_SDPL_infos (struct SDPL_map *SDPL_map)
{
#if CONFIG_BLK_SDPL_EMBEDDED
    sdpl_info("Found an embedded SDPL map version: %d.%d\n",
              (SDPL_map->magic & 0x00F0) >> 8, SDPL_map->magic & 0x000F);
#else /* !CONFIG_BLK_SDPL_EMBEDDED */
    sdpl_info("Found a%s SDPL map version: %d.%d\n",
              EMBEDDED(SDPL_map) == 0 ? "" : "n embedded",
              (SDPL_map->magic & 0x00F0) >> 8, SDPL_map->magic & 0x000F);
#endif /* !CONFIG_BLK_SDPL_EMBEDDED */
    sdpl_info("Device size: " STD_ATTRIB "\n", SDPL_map->device_size);
#if !CONFIG_BLK_SDPL_EMBEDDED
    sdpl_info("Reserved size: %llu\n", SDPL_map->res_size);
#else
    sdpl_info("Reserved size: %u\n", SDPL_map->res_size);
#endif
    sdpl_info("MBR magic: %04X - CRC: %08X\n", SDPL_map->magic, SDPL_map->CRC);
    sdpl_info("info size: %d bits\n", SDPL_map->info_size);
    sdpl_info("Bloc size: " STD_ATTRIB " sectors\n", SDPL_map->bloc_size);
    sdpl_info("Layer size: " STD_ATTRIB " blocs\n", SDPL_map->SDPL_size);
    sdpl_info("Num parts: " STD_ATTRIB "\n", NPARTS(SDPL_map));
    sdpl_info("Bitmap offset: %08lX\n",
              (unsigned long)SDPL_map->bitmap_offset);
    sdpl_info("Datas offset: %08lX\n", (unsigned long)SDPL_map->data_offset);
    sdpl_info("Partition decription offset: %08lX\n",
              (unsigned long)SDPL_map->pentries_offset);
}

void dump_part_infos (struct SDPL_map *SDPL_map, int minor)
{
    std_type bloc = 0;

    sdpl_info("Partition %c %d: Magic: %04X - len: " STD_ATTRIB
              " - offset %08lX\n",
              SDPL_map->pflags[minor] & SDPL_BOOT ? '*' : '-',
              minor + 1, (unsigned int)SDPL_map->pmagic[minor],
              SDPL_map->psize[minor], (unsigned long)SDPL_map->poffset[minor]);
    while (bloc < SDPL_map->psize[minor]) {
        std_type pbloc = SDPL_map->pbloc[minor][bloc];
        sdpl_info("| " STD_ATTRIB " - " STD_ATTRIB " ", bloc, pbloc);
        if ((SDPL_map->bflags[pbloc] & SDPL_RW) == 0)
            sdpl_info("-");
        else
            sdpl_info("*");
        bloc++;
        if ((bloc & 0x03) == 0) {
            sdpl_info("|\n");
        }
    }
    if ((bloc & 0x3) != 0)
        sdpl_info("|\n");
}

void dump_free_blocks (struct SDPL_map *SDPL_map)
{
    std_type i = 0, bloc;

    sdpl_info("Free blocs:\n");
    bloc =0;
    while (bloc < SDPL_map->SDPL_size) {
        if (SDPL_map->bflags[bloc] == EMPTY_BLOC) {
            sdpl_info("| " STD_ATTRIB " ", bloc);
            i++;
        }
        bloc++;
        if (i == 4) {
            sdpl_info("|\n");
            i = 0;
        }
    }
    if (i != 0) {
        sdpl_info("|\n");
    }
}
#endif /* CONFIG_BLK_SDPL_DEBUG > 2 || !defined(__KERNEL__) */
