/*
 * SDPL sdpl.c: main kernel code for sdpl module
 *
 *   Copyright (C) 2001 - 2007, Jocelyn MAYER <l_indien at magic dot fr>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
/*
 * Simple Disk Partition Layer
 *
 * Kernel space module
 *
 */

#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/unistd.h>
#include <linux/slab.h>
#include <linux/hdreg.h>
#include <linux/file.h>
#include <asm/uaccess.h>
#include <linux/blk.h>
/* Embedded should also mean TINY ! */
#ifdef CONFIG_PROC_FS
#include <linux/proc_fs.h>
#endif /*  CONFIG_PROC_FS */

#if !CONFIG_BLK_SDPL_EMBEDDED
#if CONFIG_UNIZLIB
#include <linux/zlib.h>
#else
#include "adler32.h"
#endif
#endif /* !CONFIG_BLK_SDPL_EMBEDDED */

#include <linux/sdpl.h>

/* Misc infos */
static char SDPL_name[] = SDPL_NAME;
#if CONFIG_BLK_SDPL_DEBUG > 1
static char SDPL_version[] = SDPL_VERSION_STR;
#endif /* CONFIG_BLK_SDPL_DEBUG > 1 */
/* Main Linux structures */
static int SDPL_blocksizes[MAX_MEDIUM * (MAX_MINOR + 1)] = {0,};
static int SDPL_sectsizes[MAX_MEDIUM * (MAX_MINOR + 1)] = {0,};
static int SDPL_sizes[MAX_MEDIUM * (MAX_MINOR + 1)] = {0,};

/* Internal structures */
 /* Embedded layer has only one medium */
#if CONFIG_BLK_SDPL_EMBEDDED
static SDPL_device_t SDPL_devs[MAX_MINOR];
static layer_t SDPL_layer = { SDPL_devs, 0, NULL, SPIN_LOCK_UNLOCKED, 0, 0 };
#else /* CONFIG_BLK_SDPL_EMBEDDED */
static layer_t SDPL_layers[MAX_MEDIUM];
/* Stats */
#if CONFIG_BLK_SDPL_STATS
static std_type SDPL_medium = 0;
#endif /* CONFIG_BLK_SDPL_STATS */
#endif /* CONFIG_BLK_SDPL_EMBEDDED */
static spinlock_t SDPL_global_lock = SPIN_LOCK_UNLOCKED;

/* Prototypes */
static int SDPL_domap (kdev_t device, int medium);
/* Block device operations */
static int SDPL_blk_open(struct inode *inode, struct file *file);
static int SDPL_blk_ioctl(struct inode *inode, struct file *file,
                          uint command, ulong a);
static int SDPL_blk_close(struct inode *inode, struct file *file);
static int SDPL_make_request(request_queue_t *q, int rw,
                             struct buffer_head *bh);

static char SDPL_gotmap = 0;
static char SDPL_mapname[64];

#ifdef MODULE
/* XXX: find a better solution */
kdev_t __init name_to_kdev_t(char *line)
{
    if (!strcmp(line, "/dev/nftla")) {
        return to_kdev_t(0x3e00);
    } else {
        return to_kdev_t(0);
    }
}
#endif

/** Preinit routine: parse arguments
 * @param str string containing arguments
 */
static int __init sdpl_set_map (char *str)
{
    spin_lock(&SDPL_global_lock);
    sdpl_debug("sdpl_set_map\n");
    strcpy(SDPL_mapname, str);
    SDPL_gotmap = 1;
    sdpl_debug("Will map device: %s\n", SDPL_mapname);
    spin_unlock(&SDPL_global_lock);
    return 1;
}

/** Helper routine to map a block device a boot time
 */
static void sdpl_open_fs (void)
{
    struct block_device *bdev;
    kdev_t layer;
    int ret = 0;
    int mode;

    if (SDPL_gotmap == 1) {
        sdpl_debug("sdpl_open_fs: Got to map device: %s\n", SDPL_mapname);
        layer = name_to_kdev_t(SDPL_mapname);
        bdev = bdget(kdev_t_to_nr(layer));
        mode = FMODE_READ | FMODE_WRITE;
        blkdev_get(bdev, mode, 0, BDEV_FS);
        sdpl_debug("sdpl_open_fs: Got device: %d - %d\n",
                   MAJOR(layer), MINOR(layer));
        ret = SDPL_domap(layer, 0);
        if (ret == 0)
            SDPL_gotmap = 2;
#if 0 /* Causes BUG ? */
        else
            blkdev_put(bdev, BDEV_FS);
#endif
    }
}

/** Last routine: release boot device
 */
static void sdpl_free_map (void)
{
    if (SDPL_gotmap == 2) {
        struct block_device *bdev;
        kdev_t layer;

        layer = name_to_kdev_t(SDPL_mapname);
        bdev = bdget(kdev_t_to_nr(layer));
        blkdev_put(bdev, BDEV_FS);
    }
}

#ifdef MODULE

static char *sdpl_map;

MODULE_PARM(sdpl_map, "s");

#if !CONFIG_BLK_SDPL_EMBEDDED
MODULE_PARM_DESC(sdpl_map, "Define a layer to map at boot time");
#endif /* !CONFIG_BLK_SDPL_EMBEDDED */

#else

__setup("sdpl_map=", sdpl_set_map);

#endif /* !MODULE */

/* block device operations structure */
static struct block_device_operations SDPL_fops =
{
#if (LINUX_VERSION_CODE >= 0x20414)
        owner:          THIS_MODULE,
#endif /* (LINUX_VERSION_CODE >= 0x20414) */
        open:           SDPL_blk_open,
        release:        SDPL_blk_close,
        ioctl:          SDPL_blk_ioctl,
};

/********************************************************************
 *
 * SDPL devfs interface
 *
 ********************************************************************/
#ifdef CONFIG_DEVFS_FS

static devfs_handle_t devfs_handle;

static int register_devfs (void)
{
    /* Register into devfs (so cool...) */
    if (devfs_register_blkdev(SDPL_MAJOR, SDPL_name, &SDPL_fops))
        return -EIO;
    devfs_handle = devfs_mk_dir(NULL, SDPL_DEVNAME, NULL);
    devfs_register_series(devfs_handle, "%u", (MAX_MINOR + 1) * MAX_MEDIUM,
                          DEVFS_FL_DEFAULT, SDPL_MAJOR, 0,
                          S_IFBLK | S_IRUSR | S_IWUSR | S_IRGRP,
                          &SDPL_fops, NULL);
    return 0;
}

static void cleanup_devfs (void)
{
    /* Deregister from devfs */
    if (devfs_unregister_blkdev(SDPL_MAJOR, SDPL_DEVNAME)) {
        sdpl_error("sdpl: cannot unregister devfs blkdev\n");
    }
    devfs_unregister(devfs_handle);
}

#endif /* CONFIG_DEVFS_FS */

/********************************************************************
 *
 * SDPL procfs interface
 *
 ********************************************************************/
#ifdef CONFIG_PROC_FS

static struct proc_dir_entry *proc_sdpl_root = NULL;

int sdpl_get_layers(char *page, char **start, off_t off,
                    int count, int *eof, void *data);
static int sdpl_get_parts(char *page, char **start, off_t off,
                          int count, int *eof, void *data);

static int register_procfs (void)
{
    /* Register into procfs */
    if ((proc_sdpl_root = proc_mkdir(SDPL_DEVNAME, 0)) == NULL)
        return -EFAULT;
    create_proc_read_entry("layers", 0, proc_sdpl_root, sdpl_get_layers, NULL);

    return 0;
}

static void sdpl_add_entry (int medium, int minor)
{
    /* For each device, create a file in /proc/sdpl/ */
    sprintf(SDPL_DEV(medium, minor).proc_name, "sdpl%c%02d",
            'a' + medium, minor);
    sdpl_debug(" %s\n", SDPL_DEV(medium, minor).proc_name);
    create_proc_read_entry(SDPL_DEV(medium, minor).proc_name, 0,
                           proc_sdpl_root, sdpl_get_parts,
                           &SDPL_DEV(medium, minor));
}

static void cleanup_procfs (void)
{
    /* Deregister from procfs */
    remove_proc_entry("sdpl/layers", 0);
    remove_proc_entry(SDPL_DEVNAME, 0);
}

/** Common routine for /proc output
 *  From linux/fs/proc/proc_misc.c
 */
static int proc_calc_metrics(char *page, char **start, off_t off,
                             int count, int *eof, int len)
{
    if (len <= off+count && eof != NULL)
        *eof = 1;
    if (start != NULL)
        *start = page + off;
    len -= off;
    if (len > count)
        len = count;
    if (len < 0)
        len = 0;
    return len;
}

/** Dump informations about mapped layers into /proc/sdpl/layers
 */
int sdpl_get_layers(char *page, char **start, off_t off,
                           int count, int *eof, void *data)
{
    struct SDPL_map *SDPL_map = NULL;
    char *pos = page;
    int medium;
    int len;
    spinlock_t *medium_lock;

    sdpl_debug("sdpl_get_layers\n");
    len = sprintf(pos, "********************* SDPL V%02u.%02u ***************"
                  "******\n", SDPL_VERSION >> 16, SDPL_VERSION & 0xFF);
    len += sprintf(pos + len,
                  "********************* SDPL Layers *********************\n");
    /* Make sure there won't be any changes while we're here */
    FOR_EACH_MEDIUM /* No semi-colum here, please !!! */
    {
        if (IS_DEV_ALLOC(medium))
        {
            medium_lock = &SDPL_LAYER(medium)->medium_lock;
            spin_lock(medium_lock);
            SDPL_map = SDPL_LAYER(medium)->SDPL_map;
            if (SDPL_map != NULL) {
                len += sprintf(pos + len, "SDPL map version: %d.%d on device "
                               "%d-%d\n", (SDPL_map->magic & 0x00F0) >> 8,
                               SDPL_map->magic & 0x000F,
                               MAJOR(SDPL_LAYER(medium)->device),
                               MINOR(SDPL_LAYER(medium)->device));
                len += sprintf(pos + len, "MBR magic: %04X - CRC: %08X\n",
                               SDPL_map->magic, SDPL_map->CRC);
                len += sprintf(pos + len, "info size: %d bits\n",
                               SDPL_map->info_size);
                len += sprintf(pos + len, "Bloc size: " STD_ATTRIB
                               " sectors\n", SDPL_map->bloc_size);
                len += sprintf(pos + len, "Layer size: " STD_ATTRIB
                               " blocs\n", SDPL_map->SDPL_size);
                len += sprintf(pos + len, "Num parts: " STD_ATTRIB "\n",
                               NPARTS(SDPL_map));
                len += sprintf(pos + len, "Bitmap offset: %08X\n",
                               (unsigned int)SDPL_map->bitmap_offset);
                len += sprintf(pos + len, "Partition decription offset: "
                               "%08X\n",
                               (unsigned int)SDPL_map->pentries_offset);
                len += sprintf(pos + len, STD_ATTRIB " free blocs\n",
                               count_free_blocs(SDPL_map));
                len += sprintf(pos + len,
                               "----------------------------"
                               "---------------------------\n");
            }
            spin_unlock(medium_lock);
        }
    }
    return proc_calc_metrics(page, start, off, count, eof, len);
}

/** Dump informations about a sdpl partition
 */
static int sdpl_get_parts(char *page, char **start, off_t off,
                   int count, int *eof, void *data)
{
    SDPL_device_t *SDPL_dev = (SDPL_device_t *)data;
    struct SDPL_map *SDPL_map;
    char *pos = page;
    int minor, bloc;
    int len;

    sdpl_debug("sdpl_get_parts\n");
    len = sprintf(pos, "********************* SDPL V%02u.%02u ***************"
                  "******\n", SDPL_VERSION >> 16, SDPL_VERSION & 0xFF);
    len += sprintf(pos + len,
                  "******************** SDPL Partition *******************\n");
    spin_lock(&(SDPL_dev->lock));
    if (SDPL_dev != NULL && SDPL_dev->layer != NULL &&
        (SDPL_map = SDPL_dev->layer->SDPL_map) != NULL) {
        minor = SDPL_dev->map_entry;
        len += sprintf(pos + len, "Partition %d: Magic: %04X - len: "
                       STD_ATTRIB " - offset %08X\n", minor,
                       (unsigned int)SDPL_map->pmagic[minor],
                       SDPL_map->psize[minor],
                       (unsigned int)SDPL_map->poffset[minor]);
        if (SDPL_map->pbloc[minor] != NULL) {
            bloc = 0;
            while (bloc < SDPL_map->psize[minor]) {
                len += sprintf(pos + len, "| %02d - " STD_ATTRIB " ",
                               bloc, SDPL_map->pbloc[minor][bloc]);
                if (SDPL_map->bflags[SDPL_map->pbloc[minor][bloc]] == SDPL_RO)
                    len += sprintf(pos + len, "-");
                else
                    len += sprintf(pos + len, "*");
                bloc++;
                if ((bloc & 0x03) == 0) {
                    len += sprintf(pos + len, "|\n");
                }
            }
            if ((bloc & 0x3) != 0)
                len += sprintf(pos + len, "|\n");
        }
    }
#if CONFIG_BLK_SDPL_STATS
    len += sprintf(pos + len, "Opens: %lu, Reads: %lu, Writes: %lu\n",
                   SDPL_dev->opens, SDPL_dev->reads, SDPL_dev->writes);
#endif /* CONFIG_BLK_SDPL_STATS */
    spin_unlock(&(SDPL_dev->lock));
    len += sprintf(pos + len, "-----------------------"
                   "--------------------------------\n");

    return proc_calc_metrics(page, start, off, count, eof, len);
}

#ifndef MODULE
EXPORT_SYMBOL(sdpl_get_layers);
#endif

#endif /* CONFIG_PROC_FS */

/********************************************************************
 *
 * SDPL internal functions
 *
 ********************************************************************/

/** Checks factorisation
 */
static int SDPL_common_checks (int medium, int minor, int mode)
{
    int ret = -ENODEV;

    sdpl_debug("SDPL_common_checks\n");
    /* Check medium / minor overflow */
#if !CONFIG_BLK_SDPL_EMBEDDED
    sdpl_debug("Check medium %d - minor %d\n", medium, minor);
    if (medium >= MAX_MEDIUM || minor >= MAX_MINOR) {
        sdpl_error("SDPL medium %d - minor %d > Max SDPL devices "
                   "(%d - %d)\n", medium, minor, MAX_MEDIUM, MAX_MINOR);
        goto checks_end;
    }
#endif /* !CONFIG_BLK_SDPL_EMBEDDED */
    sdpl_debug("Check minor %d - max " STD_ATTRIB "\n", minor,
               SDPL_LAYER(medium)->max);
    if (minor >= SDPL_LAYER(medium)->max) {
        sdpl_error("SDPL minor %d > number of partition of the device"
                   "(%d - " STD_ATTRIB ")\n", minor, medium,
                   SDPL_LAYER(medium)->max);
        goto checks_end;
    }
    /* Check for configgured SDPL partition */
    sdpl_debug("Check device & map\n");
    if (!IS_DEV_ALLOC(medium) ||
        SDPL_LAYER(medium)->SDPL_map == NULL) {
        sdpl_error("No SDPL device minor %d configured\n", minor);
        goto checks_end;
    }
    /* Check for access mode */
    sdpl_debug("Check access mode\n");
    if ((mode & FMODE_WRITE) && !(SDPL_DEV(medium, minor).access & SDPL_RW)) {
        sdpl_error("RW access to RO SDPL device %d\n", minor);
        ret = -EACCES;
        goto checks_end;
    }
    ret = 0;

 checks_end:
    sdpl_debug("SDPL_common_checks end\n");
    return ret;
}

/** Generic routine to flush one device's caches
 */
static void do_flush (kdev_t device)
{
    /* Change it so it act like blkpg.c in recent kernels */
    /* Make sure sync won't return before sync has been done */
    fsync_dev(device);
    invalidate_buffers(device);
}

/** Free structures about a layer
 */
static void free_device (struct SDPL_map *SDPL_map, int medium)
{
    sdpl_debug("free_device\n");
    if (SDPL_map != NULL) {
        sdpl_free(SDPL_map->bflags);
        sdpl_free(SDPL_map->pbloc);
        sdpl_free(SDPL_map->pnext);
        sdpl_free(SDPL_map->poffset);
        sdpl_free(SDPL_map->psize);
        sdpl_free(SDPL_map->pflags);
        sdpl_free(SDPL_map->pmagic);
        sdpl_free(SDPL_map->pentries);
        sdpl_free(SDPL_map);
    }
    FREE_DEV(medium);
    sdpl_debug("nulify device\n");
    SDPL_sizes[medium << MEDIUM_SHIFT] = 0;
    SDPL_blocksizes[medium << MEDIUM_SHIFT] = 0;
    SDPL_sectsizes[medium << MEDIUM_SHIFT] = 0;
#if CONFIG_BLK_SDPL_STATS
    SDPL_medium--;
#endif /* CONFIG_BLK_SDPL_STATS */
}

/** Free structures about one sdpl partition
 */
static void free_partition (int medium, int minor)
{
    struct SDPL_map *SDPL_map = SDPL_LAYER(medium)->SDPL_map;
    spinlock_t *minor_lock = &SDPL_DEV(medium, minor).lock;

    sdpl_debug("free_partition\n");
    /* Do whatever is needed (free structures...) */
    spin_lock(minor_lock);
    if (SDPL_map->pbloc[minor] != NULL) {
        sdpl_free(SDPL_map->pbloc[minor]);
    }
    SDPL_DEV(medium, minor).map_entry = 0;
    MOD_INIT_READS(medium, minor);
    MOD_INIT_WRITES(medium, minor);
    SDPL_DEV(medium, minor).access = SDPL_RW;
    SDPL_sizes[(medium << MEDIUM_SHIFT) + minor + 1] = 0;
    SDPL_blocksizes[(medium << MEDIUM_SHIFT) + minor + 1] = 0;
    SDPL_sectsizes[(medium << MEDIUM_SHIFT) + minor + 1] = 0;
    REMOVE_PROC_ENTRY(medium, minor);
    spin_unlock(minor_lock);
    SDPL_LAYER(medium)->minors--;
}

/** Map a new layer
 */
static int SDPL_domap (kdev_t device, int medium)
{
    Sector sect;
    block_dev_t *bdev;
    struct SDPL_map *SDPL_map;
    int minor;
    int ret = -EINVAL;

    sdpl_debug("SDPL_domap\n");
    if (IS_DEV_ALLOC(medium)) {
        sdpl_error("Device already mapped\n");
        return -EINVAL;
    }
#if !CONFIG_BLK_SDPL_EMBEDDED
    /* Check that we don't map a layer inside itself (would deadlock) */
    if (MAJOR(device) == SDPL_MAJOR &&
        (MINOR(device) >> MEDIUM_SHIFT) == medium &&
        (MINOR(device) & MASK_MINOR) == 0) {
        sdpl_error("Trying to map a layer recursively\n");
        return -EINVAL;
    }
#endif
    if (blk_size[MAJOR(device)] == NULL) {
        sdpl_error("Device %02x:%02x is not registered\n",
                   MAJOR(device), MINOR(device));
        return -ENODEV;
    }

    if ((SDPL_map = sdpl_malloc(sizeof(struct SDPL_map))) == NULL) {
        sdpl_error("Memory error\n");
        return -ENOMEM;
    }
    SDPL_map->device_size = blk_size[MAJOR(device)][MINOR(device)];
    SDPL_map->infos = NULL;
    SDPL_map->pentries = NULL;
    SDPL_map->pmagic = NULL;
    SDPL_map->pflags = NULL;
    SDPL_map->psize = NULL;
    SDPL_map->poffset = NULL;
    SDPL_map->pnext = NULL;
    SDPL_map->pbloc = NULL;
    SDPL_map->bflags = NULL;

    /* Get the first sector from the device */
    ALLOC_BLOCKDEV(bdev, device);
    sdpl_debug("Try read sector\n");
    if (bdev == NULL ||
        (SDPL_map->mbr =
         (SDPL_br_t *)read_dev_sector(bdev, 0, &sect)) == NULL) {
        sdpl_error("Can't access device...\n");
        sdpl_free(SDPL_map);
        return -ENODEV;
    }
    sdpl_debug("Try get infos\n");
    if (get_SDPL_infos(SDPL_map) < 0) {
        sdpl_error("Can't get sdpl infos\n");
        goto domap_freemap;
    }
#if !CONFIG_BLK_SDPL_EMBEDDED
    if ((SDPL_LAYER(medium)->SDPL_dev =
         sdpl_malloc(sizeof(SDPL_device_t) * NPARTS(SDPL_map))) == NULL) {
        sdpl_error("Not enough memory\n");
        ret = -ENOMEM;
        goto domap_error;
    }
#endif /* !CONFIG_BLK_SDPL_EMBEDDED */
    sdpl_debug("Got infos from map: %08X\n", (unsigned int)SDPL_map);
#if CONFIG_BLK_SDPL_DEBUG > 2
    dump_SDPL_infos(SDPL_map);
#endif /* CONFIG_BLK_SDPL_DEBUG > 2 */

    /* Should check that we won't be over MAX_MINOR */
    if (NPARTS(SDPL_map) > MAX_MINOR) {
        sdpl_error("Too many SDPL partitions (" STD_ATTRIB " - " STD_ATTRIB
                   ")\n", NPARTS(SDPL_map), (std_type)MAX_MINOR);
        goto domap_error;
    }
    sdpl_debug("Get parts\n");
    get_parts_infos(SDPL_map, sect, bdev);

    SDPL_LAYER(medium)->SDPL_map = SDPL_map;
    SDPL_LAYER(medium)->device = device;
    for (minor = 0; minor < NPARTS(SDPL_map); minor++) {
#if CONFIG_BLK_SDPL_DEBUG > 2
        dump_part_infos(SDPL_map, minor);
#endif /* CONFIG_BLK_SDPL_DEBUG > 2 */
        SDPL_DEV(medium, minor).layer = SDPL_LAYER(medium);
        /* We got to find NPARTS free entries... */
        sdpl_debug("Attach map %p entry %02d\n",
                   SDPL_map, minor);
        SDPL_DEV(medium, minor).map_entry = minor;
        MOD_INIT_READS(medium, minor);
        MOD_INIT_WRITES(medium, minor);
        MOD_INIT_OPENS(medium, minor);
        SDPL_DEV(medium, minor).access =
            (SDPL_map->pflags[minor] & SDPL_RW) ? SDPL_RW : SDPL_RO;
        SDPL_DEV(medium, minor).lock = SPIN_LOCK_UNLOCKED;
        /* TO FIX */
#if 0
        SDPL_blocksizes[(medium << MEDIUM_SHIFT) + minor + 1] =
            SDPL_map->bloc_size * 512;
        SDPL_sectsizes[(medium << MEDIUM_SHIFT) + minor + 1] =
            512;
#endif
        SDPL_sizes[(medium << MEDIUM_SHIFT) + minor + 1] =
            (SDPL_map->psize[minor] * SDPL_map->bloc_size *
             512) >> BLOCK_SIZE_BITS;
        ADD_PROC_ENTRY(medium, minor);
    }
#if CONFIG_BLK_SDPL_DEBUG > 2
    dump_free_blocks(SDPL_map);
#endif /* CONFIG_BLK_SDPL_DEBUG > 2 */
#if !CONFIG_BLK_SDPL_EMBEDDED
    sdpl_debug("Fix layer size\n");
#ifdef CONFIG_BLK_DEV_LOOP
    if (blksize_size[MAJOR(device)][MINOR(device)] == 0) {
        /* Fix for loop device use... */
        SDPL_sizes[medium << MEDIUM_SHIFT] = SDPL_map->device_size;
    } else
#endif
    {
        /* Fix for ramdisk use */
#ifdef CONFIG_BLK_DEV_RAM
        if (MAJOR(device) == RAMDISK_MAJOR) {
            SDPL_sizes[medium << MEDIUM_SHIFT] =
                SDPL_map->device_size /
                blksize_size[MAJOR(device)][MINOR(device)];
        } else
#endif
        {
            SDPL_sizes[medium << MEDIUM_SHIFT] = SDPL_map->device_size;
        }
    }
#endif /* !CONFIG_BLK_SDPL_EMBEDDED */
    /* TO FIX */
#if 0
    SDPL_blocksizes[medium << MEDIUM_SHIFT] = SDPL_map->bloc_size * 512;
    SDPL_sectsizes[medium << MEDIUM_SHIFT] = 512;
#endif
    sdpl_debug("Fix sdpl infos\n");
    SDPL_LAYER(medium)->minors = NPARTS(SDPL_map);
    SDPL_LAYER(medium)->max = NPARTS(SDPL_map);
#if CONFIG_BLK_SDPL_STATS
    SDPL_medium++;
    sdpl_debug("After map, " STD_ATTRIB " sdpl device are in use\n",
               SDPL_medium);
#endif /* CONFIG_BLK_SDPL_STATS */

    ret = 0;

 domap_end:
    sdpl_debug("Put_dev_sector\n");
#if !CONFIG_BLK_SDPL_EMBEDDED
    put_dev_sector(sect);
#endif
    sdpl_debug("sdpl_domap returns %d\n", ret);
    return ret;

 domap_error:
    sdpl_debug("Error in sdpl_domap\n");
    FREE_DEV(medium);
 domap_freemap:
    sdpl_free(SDPL_map);
    goto domap_end;
}

/** Routine to unmap a partition
 */
static int SDPL_dounmap (struct inode *inode, int medium, int minor)
{
    struct SDPL_map *SDPL_map;
    kdev_t device;

    sdpl_debug("SDPL_dounmap\n");
    /* First, invalidate all caches */
    device = inode->i_rdev;

    sdpl_debug("About to unmap sdpl device: %d\n", minor);
    /* Should already have been detected... */
    if (!IS_DEV_ALLOC(medium) ||
        (SDPL_map = SDPL_LAYER(medium)->SDPL_map) == NULL) {
        sdpl_error("Device not mapped\n");
        return -ENODEV;
    }
    /* We got to free the map if it is the last entry to be unmapped */
    free_partition(medium, minor);
    if (SDPL_LAYER(medium)->minors == 0) {
        free_device(SDPL_map, medium);
        SDPL_LAYER(medium)->max = 0;
    }
#if CONFIG_BLK_SDPL_STATS
    sdpl_debug("After unmap, "STD_ATTRIB " sdpl device are in use\n",
               SDPL_medium);
#endif /* CONFIG_BLK_SDPL_STATS */

    return 0;
}

static int commit_sdpl_infos (int medium)
{
    Sector sector;
    std_type cur_sect = 1;
    block_dev_t *bdev;

    sdpl_info("Going to flush device\n");
    ALLOC_BLOCKDEV(bdev, SDPL_LAYER(medium)->device);
    /*
     * We ALWAYS start with second sector
     * (which is a copy of the MBR
     */
    if (read_dev_sector(bdev, cur_sect, &sector) < 0)
        return -1;

    save_sdpl(SDPL_LAYER(medium)->SDPL_map, SDPL_LAYER(medium)->device,
              &sector, &cur_sect);
    sdpl_info("Flush device\n");
    do_flush(SDPL_LAYER(medium)->device);
    RELEASE_BLOCKDEV(bdev);

    return 0;
}

/********************************************************************
 *
 * Block device functions
 *
 ********************************************************************/
/** block device open routine
 */
/* I comment MOD_INC/DEC_USE_COUNT for now,
 *  so I can test more easily (bad way of doing !!!)
 */
static int SDPL_blk_open(struct inode *inode, struct file *file)
{
    int medium, minor;
    int ret = -1;
    spinlock_t *medium_lock;
    spinlock_t *minor_lock;

    sdpl_debug("SDPL_blk_open\n");
    /* This should never happen (Kernel Bug ?) */
    if (INVALID_REQUEST(inode, inode->i_rdev)) {
        sdpl_error("Invalid open...\n");
        return -EINVAL;
    }

    sdpl_debug("Try open\n");
    spin_lock(&SDPL_global_lock);
    sdpl_open_fs();
    minor = MINOR(inode->i_rdev);
    medium = minor >> MEDIUM_SHIFT;
    minor = minor & MASK_MINOR;
    medium_lock = &SDPL_LAYER(medium)->medium_lock;
    minor_lock = &SDPL_DEV(medium, minor).lock;
    sdpl_info("About to open device: %d - %d\n", medium, minor);
    spin_lock(medium_lock);
    spin_unlock(&SDPL_global_lock);
    if (minor > 0) {
        minor--;
        sdpl_debug("Got to check minor %d\n", minor);
        /* Open a partition */
        if (file == NULL)
            ret = SDPL_common_checks(medium, minor, 0);
        else
            ret = SDPL_common_checks(medium, minor, file->f_mode);
        spin_lock(minor_lock);
        spin_unlock(medium_lock);
        if (ret == 0) {
            /* All right */
            MOD_INC_OPENS(medium, minor);
//            MOD_INC_USE_COUNT;
        }
        spin_unlock(minor_lock);
    } else {
        /* Opens on "medium" devices, ie open a layer */
        if (minor < 0 || medium > MAX_MEDIUM) {
            sdpl_error("Invalid medium/minor %d/%d (%d - %d)\n",
                       medium, minor, MAX_MEDIUM, MAX_MINOR);
            ret = -ENODEV;
        } else {
            sdpl_info("Open medium: %d\n", medium);
//        MOD_INC_USE_COUNT;
            ret = 0;
        }
        spin_unlock(medium_lock);
    }
    sdpl_debug("SDPL_blk_open: end\n");

    return ret;
}

/** block device i/o-control routine on layers
 */
static int SDPL_medium_ioctl(kdev_t dev_rdev, uint command, ulong a,
                             int medium)
{
    struct SDPL_map *SDPL_map = SDPL_LAYER(medium)->SDPL_map;
    int ret = -EINVAL;

    sdpl_debug("IOCTL on medium %d\n", medium);
    //    sdpl_info("IOCTL on medium %d\n", medium);
    switch (command) {
    case SDPL_ATTACH:
        do {
            struct file *dev_file;
            struct inode *dev_inode;
            kdev_t device;

            sdpl_debug("%s -- SDPL_ATTACH\n", SDPL_name);
            CHECK_PERMS;
            if ((dev_file = fget(a)) == NULL) {
                sdpl_error("Bad file...\n");
                break;
            }
            if ((dev_inode = dev_file->f_dentry->d_inode) == NULL) {
                sdpl_error("Null inode.\n");
                break;
            }
            if (!S_ISBLK(dev_inode->i_mode)) {
                sdpl_error("Not a block device !\n");
                break;
            }
            device = dev_inode->i_rdev;
            do_flush(device);
            ret = SDPL_domap(device, medium);
            sdpl_debug("IOCTL: got %d from domap\n", ret);
            fput(dev_file);
            break;
        } while (0);
        break;

    case BLKFLSBUF:
        /* flush buffer cache */
        sdpl_debug("%s -- BLKFLSBUF\n", SDPL_name);
        do {
            CHECK_PERMS;
            commit_sdpl_infos(medium);
            do_flush(SDPL_LAYER(medium)->device);
            //            do_flush(dev_rdev);
            ret = 0;
        } while (0);
        break;

    case SDPL_SET_BLOC_ACCESS:
        sdpl_debug("%s -- SDPL_SET_BLK_ACCESS\n", SDPL_name);
    {
        std_type blocnum;
        /* set access flags of a logical volume */
        CHECK_PERMS;
        /* Flush, it won't be possible if we get read-only... */
        do_flush(SDPL_LAYER(medium)->device);
        if (copy_from_user(&blocnum, (char *)a, sizeof(std_type)) < 0) {
            ret =  -EFAULT;
            break;
        }
        toggle_bloc_flag(SDPL_map, blocnum, -1);
        commit_sdpl_infos(medium);
        ret = 0;
        break;
    }

    case SDPL_GET_BLOC_ACCESS:
        sdpl_debug("%s -- SDPL_GET_BLK_ACCESS\n", SDPL_name);
        {
            struct SDPL_map *SDPL_map = SDPL_LAYER(medium)->SDPL_map;
            std_type blocnum;

            if (copy_from_user(&blocnum, (char *)a, sizeof(std_type)) < 0) {
                ret =  -EFAULT;
                break;
            }
            sdpl_info("GET_ACCESS blocnum: " STD_ATTRIB " - %d\n",
                      blocnum, SDPL_map->bflags[blocnum]);
            if (put_user(SDPL_map->bflags[blocnum], (u16 *)a))
                ret = -EFAULT;
            else
                ret = 0;
        }
        break;

    case BLKGETSIZE:
#if CONFIG_BLK_SDPL_64BITS
    case BLKGETSIZE64:
#endif /* CONFIG_BLK_SDPL_64BITS */
        /* return device size */
        {
            /* BEWARE: this is an int to respect the kernel API */
            int devsize = SDPL_map->device_size * 2;
            sdpl_debug("%s -- BLKGETSIZE: " STD_ATTRIB "\n",
                   SDPL_name, SDPL_map->device_size);
            if (copy_to_user((int *)a, &devsize,
                             sizeof(int)) != 0)
                ret = -EFAULT;
            else
                ret = 0;
        }
        break;

    case SDPL_GET_BLOCSIZE:
        sdpl_debug("%s -- SDPL_GET_BLOCSIZE\n", SDPL_name);
        if (put_user(SDPL_map->bloc_size, (std_type *)a))
            ret = -EFAULT;
        else
            ret = 0;
        break;

    case SDPL_GET_RESSIZE:
        sdpl_debug("%s -- SDPL_GET_RESSIZE " STD_ATTRIB "\n",
                   SDPL_name, SDPL_map->res_size);
        if (put_user(SDPL_map->res_size, (std_type *)a))
            ret = -EFAULT;
        else
            ret = 0;
        break;

    case SDPL_GET_NUMPARTS:
        sdpl_debug("%s -- SDPL_GET_NUMPARTS" STD_ATTRIB "\n",
                   SDPL_name, SDPL_LAYER(medium)->minors);
        if (put_user(SDPL_LAYER(medium)->minors, (std_type *)a))
            ret = -EFAULT;
        else
            ret = 0;
        break;

    default:
        sdpl_error("%s ioctl: unknown command %x\n",
                   SDPL_name, command);
        break;
    }

    return ret;
}

static int SDPL_minor_ioctl(uint command, void *arg, int medium, uint minor)
{
    std_type SDPL_spb, SDPL_size, dev_size;
    struct SDPL_map *SDPL_map;
    int ret = -EINVAL;

    /* Get infos */
    SDPL_map = SDPL_LAYER(medium)->SDPL_map;
    SDPL_spb = SDPL_map->bloc_size;
    SDPL_size = SDPL_map->psize[minor];
    dev_size = SDPL_size * SDPL_spb * 512;
    switch (command) {

    case BLKGETSIZE:
#if CONFIG_BLK_SDPL_64BITS
    case BLKGETSIZE64:
#endif /* CONFIG_BLK_SDPL_64BITS */
        /* return device size */
        {
            int devsize = SDPL_size * SDPL_spb;
            sdpl_debug("%s -- BLKGETSIZE: " STD_ATTRIB "\n",
                       SDPL_name, SDPL_size);
            if (copy_to_user((int *)arg, &devsize,
                             sizeof(int)) != 0)
                ret = -EFAULT;
            else
                ret = 0;
        }
        break;

    case BLKFLSBUF:
        /* flush buffer cache */
        sdpl_debug("%s -- BLKFLSBUF\n", SDPL_name);
        CHECK_PERMS;
        do_flush(SDPL_LAYER(medium)->device);
        ret = 0;
        break;

#if !CONFIG_BLK_SDPL_EMBEDDED
    case BLKRASET:
        /* set read ahead for block device */
        sdpl_debug("%s -- BLKRAGET\n", SDPL_name);
        CHECK_PERMS;
        break;

    case BLKRAGET:
        /* get current read ahead setting */
        sdpl_debug("%s -- BLKRAGET\n", SDPL_name);
        break;

    case HDIO_GETGEO:
        {
            long start = 0;
            struct hd_geometry *hd = NULL;
            short cylinders;
            unsigned char heads = 64;
            unsigned char sectors = 32;

            /* get disk geometry - from loop device code */
            sdpl_debug("%s -- HDIO_GETGEO\n", SDPL_name);
            if (arg == NULL) {
                sdpl_error("Nowhere to put the result\n");
                break;
            }

            if (copy_from_user(hd, arg, sizeof(struct hd_geometry)) < 0) {
                ret =  -EFAULT;
                break;
            }

            cylinders = dev_size / heads / sectors;
            sdpl_debug("%s -- cylinders: " STD_ATTRIB "\n",
                       SDPL_name, dev_size / heads / sectors);
            if (copy_to_user((char *) &hd->heads, &heads,
                             sizeof(heads)) != 0 ||
                copy_to_user((char *) &hd->sectors, &sectors,
                             sizeof(sectors)) != 0 ||
                copy_to_user((short *) &hd->cylinders,
                             &cylinders, sizeof(cylinders)) != 0 ||
                copy_to_user((long *) &hd->start, &start,
                             sizeof(start)) != 0) {
                ret = -EFAULT;
                break;
            }
            ret = 0;
        }
        break;
#endif /* !CONFIG_BLK_SDPL_EMBEDDED */

    case SDPL_SET_ACCESS:
        sdpl_info("%s -- SDPL_SET_ACCESS\n", SDPL_name);
        /* set access flags of a logical volume */
        CHECK_PERMS;
        /* Flush, it won't be possible if we get read-only... */
        do_flush(SDPL_LAYER(medium)->device);
        SDPL_DEV(medium, minor).access = (ulong)arg;
        if (SDPL_DEV(medium, minor).access & SDPL_RW) {
            set_device_ro(SDPL_LAYER(medium)->device, 0);
            toggle_part_flag(SDPL_map, minor, SDPL_RO);
        } else {
            set_device_ro(SDPL_LAYER(medium)->device, 1);
            toggle_part_flag(SDPL_map, minor, SDPL_RW);
        }
        commit_sdpl_infos(medium);
        ret = 0;
        break;

    case SDPL_GET_ACCESS:
        sdpl_info("%s -- SDPL_GET_ACCESS\n", SDPL_name);
        if (copy_to_user((u16 *)arg, &(SDPL_DEV(medium, minor).access),
                         sizeof(u16)) != 0)
            ret = -EFAULT;
        else
            ret = 0;
        break;

    case SDPL_SET_BOOTABLE:
        sdpl_info("%s -- SDPL_SET_BOOTABLE %u\n",
                  SDPL_name, minor);
        mark_partition_bootable(SDPL_map, minor);
        commit_sdpl_infos(medium);
        ret = 0;
        break;

    case SDPL_GET_BOOTABLE:
        {
            std_type boot = SDPL_map->pflags[minor] & SDPL_BOOT;
            sdpl_info("%s -- SDPL_GET_BOOTABLE %u\n",
                      SDPL_name, SDPL_map->pflags[minor]);
            if (copy_to_user((u16 *)arg, &boot, sizeof(u16)) != 0)
                ret = -EFAULT;
            else
                ret = 0;
        }
        break;

    case SDPL_SET_PRIVATE_FLAGS:
        sdpl_debug("%s -- SDPL_SET_PRIVATE_FLAGS %u\n",
                   SDPL_name, minor);
    {
        int xflags;
        u8 flags;
        /* set private flags of a logical volume */
        CHECK_PERMS;
        if (copy_from_user(&xflags, (char *)arg, sizeof(int)) < 0) {
            ret =  -EFAULT;
            break;
        }
        flags = (xflags) & 0xFF;
        /*
         * Don't even check value:
         * we're supposed not to have any idea of what these flags mean !
         */
        SDPL_map->pflags[minor] = (SDPL_map->pflags[minor] & SDPL_PUBLIC) |
            (flags << SDPL_PUBLIC_BITS);
        commit_sdpl_infos(medium);
        ret = 0;
        break;
    }

    case SDPL_GET_PRIVATE_FLAGS:
        {
            int flags;
            sdpl_info("%s -- SDPL_GET_PRIVATE_FLAGS %u\n",
                      SDPL_name, SDPL_map->pflags[minor] & SDPL_PRIVATE);
            flags = (SDPL_map->pflags[minor] & SDPL_PRIVATE) >>
                    SDPL_PUBLIC_BITS;
            if (copy_to_user((int *)arg, &flags, sizeof(int)) != 0)
                ret = -EFAULT;
            else
                ret = 0;
        }
        break;

    case SDPL_FREE_BLOC:
        sdpl_info("%s -- SDPL_FREE_BLOC \n", SDPL_name);
        {
            std_type num_bloc;

            CHECK_PERMS;
            do_flush(SDPL_LAYER(medium)->device);
            if (copy_from_user(&num_bloc, arg, sizeof(std_type)) < 0) {
                ret = -EFAULT;
                break;
            }
            sdpl_info("Going to free bloc " STD_ATTRIB " from %u\n",
                      num_bloc, minor);
            /* Assume that the bloc is REALLY free */
            if ((ret = remove_bloc_from_part(SDPL_map, minor, num_bloc))
                >= 0) {
                SDPL_sizes[(medium << MEDIUM_SHIFT) + minor + 1] -=
                    (SDPL_map->bloc_size * 512) >> BLOCK_SIZE_BITS;
                sdpl_info("Now, size is " STD_ATTRIB " (minor: %u)\n",
                          SDPL_sizes[(medium << MEDIUM_SHIFT) + minor + 1],
                          minor);
                commit_sdpl_infos(medium);
            }
        }
        break;

    case SDPL_ADD_BLOCS:
        sdpl_info("%s -- SDPL_ADD_BLOCS\n", SDPL_name);
        {
            std_type nb_blocs;

            CHECK_PERMS;
            do_flush(SDPL_LAYER(medium)->device);
            if (copy_from_user(&nb_blocs, arg, sizeof(std_type)) < 0) {
                ret = -EFAULT;
                break;
            }
            if ((ret = add_blocs_to_part(SDPL_map, minor, nb_blocs)) >= 0) {
                SDPL_sizes[(medium << MEDIUM_SHIFT) + minor + 1] +=
                    (SDPL_map->bloc_size * 512 * nb_blocs) >> BLOCK_SIZE_BITS;
                sdpl_info("Now, size is " STD_ATTRIB " (minor: %u)\n",
                          SDPL_sizes[(medium << MEDIUM_SHIFT) + minor + 1],
                          minor);
                commit_sdpl_infos(medium);
            }
        }
        break;

    default:
        sdpl_error("%s ioctl: unknown command %x\n",
                   SDPL_name, command);
        break;
    }

    return ret;
}

/** block device i/o-control routine
 */
static int SDPL_blk_ioctl(struct inode *inode, struct file *file,
                          uint command, ulong a)
{
    kdev_t dev_rdev = inode->i_rdev;
    spinlock_t *medium_lock;
    spinlock_t *minor_lock;
    void *arg = (void *)a;
    int medium, minor;
    int ret = -EINVAL;

    sdpl_debug("SDPL_blk_ioctl\n");
    /* This should never happen (Kernel Bug ?) */
    if (INVALID_REQUEST(inode, dev_rdev)) {
        sdpl_error("Bad request...\n");
        goto end_ioctl;
    }
    /* Special case for SDPL_GET_VERSION ioctl (no lock and no check) */
    if (command == SDPL_GET_VERSION) {
        int version = SDPL_VERSION;

        if (copy_to_user((int *)arg, &version, sizeof(int)) != 0)
            ret = -EFAULT;
        else
            ret = 0;
        goto end_ioctl;
    }

    spin_lock(&SDPL_global_lock);
    /* If no device is atached, it can an attachement request... */
    minor = MINOR(dev_rdev);
    medium = minor >> MEDIUM_SHIFT;
    minor = minor & MASK_MINOR;
    medium_lock = &SDPL_LAYER(medium)->medium_lock;
    minor_lock = &SDPL_DEV(medium, minor).lock;

    sdpl_debug("%s -- blk_ioctl MEDIUM: %d - MINOR: %d  command: 0x%X"
               " arg: %X\n", SDPL_name, medium, minor, command, (uint)arg);
    spin_lock(medium_lock);
    spin_unlock(&SDPL_global_lock);

    /* Check for operations on the "medium" device */
    if (minor == 0) {
        ret = SDPL_medium_ioctl(dev_rdev, command, a, medium);
        spin_unlock(medium_lock);
        goto end_ioctl;
    }
    /* OK, we want to do something on a partition */
    sdpl_debug("IOCTL on medium - minor %d %d\n", medium, minor);
    if ((--minor) < 0  || SDPL_common_checks(medium, minor, 0) < 0) {
        sdpl_error("Check failed...\n");
        ret = -ENODEV;
        spin_unlock(medium_lock);
        goto end_ioctl;
    }

    /*
     * For detach, we need all to stay locked,
     * as we may free the entire layer...
     */
    if (command == SDPL_DETACH) {
        do {
            CHECK_PERMS;
            do_flush(SDPL_LAYER(medium)->device);
            //            do_flush(dev_rdev);
            ret = SDPL_dounmap(inode, medium, minor);
        } while (0);
        spin_unlock(medium_lock);
        goto end_ioctl;
    }

    spin_lock(minor_lock);
    spin_unlock(medium_lock);
    ret = SDPL_minor_ioctl(command, arg, medium, minor);
    spin_unlock(minor_lock);

 end_ioctl:
    sdpl_debug("sdpl_ioctl returns %d\n", ret);
    return ret;
}

/** block device close routine
 */
static int SDPL_blk_close(struct inode *inode, struct file *file)
{
    uint medium, minor;
    int ret = -1;
    spinlock_t *medium_lock;
    spinlock_t *minor_lock;

    sdpl_debug("SDPL_blk_close\n");
    //    sdpl_info("SDPL_blk_close\n");
    /* This should never happen (Kernel Bug ?) */
    if (INVALID_REQUEST(inode, inode->i_rdev)) {
        sdpl_error("Bad request...\n");
        return -EINVAL;
    }

    spin_lock(&SDPL_global_lock);
    minor = MINOR(inode->i_rdev);
    medium = minor >> MEDIUM_SHIFT;
    minor = minor & MASK_MINOR;
    medium_lock = &SDPL_LAYER(medium)->medium_lock;
    minor_lock = &SDPL_DEV(medium, minor).lock;

    sdpl_debug("About to close device: %d - %d\n", medium, minor);
    //    sdpl_info("About to close device: %d - %d\n", medium, minor);
    spin_lock(medium_lock);
    spin_unlock(&SDPL_global_lock);
    if (minor > 0) {
        /* Close a sdpl partition */
        minor--;
        sdpl_debug("Got to check minor %d\n", minor);
        //        sdpl_info("Got to check minor %d\n", minor);
        if (file == NULL)
            ret = SDPL_common_checks(medium, minor, 0);
        else
            ret = SDPL_common_checks(medium, minor, file->f_mode);
        spin_lock(minor_lock);
        spin_unlock(medium_lock);
        if (ret == 0) {
            /* All right */
            MOD_DEC_OPENS(medium, minor);
//            MOD_DEC_USE_COUNT;
        }
        spin_unlock(minor_lock);
    } else {
        /* close a sdpl layer */
        //        sdpl_info("Got to close medium device: %d\n", medium);
        if (minor < 0 || medium >= MAX_MEDIUM) {
            sdpl_error("Invalid medium/minor %d/%d (%d - %d)\n",
                       medium, minor, MAX_MEDIUM, MAX_MINOR);
            return -ENODEV;
        }
//        MOD_DEC_USE_COUNT;
        spin_unlock(medium_lock);
        ret = 0;
    }

    sdpl_debug("SDPL_blk_close end\n");
    //    sdpl_info("SDPL_blk_close end\n");
    return ret;
}

/** block device support function for /usr/src/linux/drivers/block/ll_rw_blk.c
 */
static int SDPL_make_request (request_queue_t *q, int rw,
                              struct buffer_head *bh)
{
    struct SDPL_map *SDPL_map;
    std_type SDPL_block, SDPL_spb, SDPL_size;
    std_type SDPL_entry;
    ulong rsector;
    ulong dev_block, dev_rsector;
    kdev_t dev_rdev;
    int medium, minor;
    spinlock_t *medium_lock;
    spinlock_t *minor_lock;

    sdpl_debug("SDPL_make_request\n");
    if (INVALID_REQUEST(bh, bh->b_rdev)) {
        sdpl_error("Bad request.\n");
        goto SDPL_req_failed;
    }

    spin_lock(&SDPL_global_lock);
    minor = MINOR(bh->b_rdev);
    medium = minor >> MEDIUM_SHIFT;
    minor = minor & MASK_MINOR;
    medium_lock = &SDPL_LAYER(medium)->medium_lock;
    minor_lock = &SDPL_DEV(medium, minor).lock;

    spin_lock(medium_lock);
    spin_unlock(&SDPL_global_lock);
    if ((SDPL_map = SDPL_LAYER(medium)->SDPL_map) == NULL) {
        sdpl_error("Device not mapped\n");
        goto SDPL_req_failed;
    }
    dev_rdev = SDPL_LAYER(medium)->device;
    if (minor > 0) {
        /* Request into a sdpl partition */
        minor--;
        if (SDPL_common_checks(medium, minor, 0) < 0) {
            sdpl_error("Device not configured\n");
            spin_unlock(medium_lock);
            goto SDPL_req_failed;
        }
        spin_lock(minor_lock);
        spin_unlock(medium_lock);
        /* Get infos */
        SDPL_entry = SDPL_DEV(medium, minor).map_entry;
        /* Sectors per block */
        SDPL_spb = SDPL_map->bloc_size;
        /* Block concerned */
        rsector = bh->b_rsector;
#if !CONFIG_BLK_SDPL_EMBEDDED || 1
        if (SDPL_map->bloc_dec == 0)
            SDPL_block = rsector / SDPL_spb;
        else
            SDPL_block = rsector >> SDPL_map->bloc_dec;
#else
        SDPL_block = rsector >> SDPL_map->bloc_dec;
#endif
        SDPL_size = SDPL_map->psize[SDPL_entry];
        dev_block = SDPL_map->pbloc[SDPL_entry][SDPL_block];

        sdpl_debug("Try to access sector %ld of entry " STD_ATTRIB
                   " -> " STD_ATTRIB "\n", rsector, SDPL_entry, SDPL_block);
#if 0
        sdpl_info("Try to access sector %ld of entry " STD_ATTRIB "\n",
                  rsector, SDPL_entry);
#endif
        /* Prevent overlapping on other partitions */
        if (SDPL_block >= SDPL_size) {
            sdpl_error("SDPL: try to access beyond end of device ("
                       STD_ATTRIB " - " STD_ATTRIB ")\n", SDPL_block,
                       SDPL_size);
            spin_unlock(minor_lock);
            goto SDPL_req_failed;
        }
        /* Then, map bloc in device */
#if !CONFIG_BLK_SDPL_EMBEDDED || 1
        if (SDPL_map->bloc_dec == 0) {
            /* Get offset in SDPL block */
            rsector -= (SDPL_block * SDPL_spb);
            dev_rsector = (dev_block * SDPL_spb) + rsector +
                (SDPL_map->data_offset >> 9);
        } else {
            rsector -= (SDPL_block << SDPL_map->bloc_dec);
            dev_rsector = (dev_block << SDPL_map->bloc_dec) + rsector +
                (SDPL_map->data_offset >> 9);
        }
#else
        rsector -= (SDPL_block << SDPL_map->bloc_dec);
        dev_rsector = (dev_block << SDPL_map->bloc_dec) + rsector +
                (SDPL_map->data_offset >> 9);
#endif
        /* statistic */
#if 0
        if (rw == WRITE || rw == WRITEA) {
            if (!(SDPL_map->bflags[SDPL_block] & SDPL_RW)) {
                sdpl_error("Try to write a read-only bloc !(" STD_ATTRIB
                           ")\n", SDPL_block);
                spin_unlock(minor_lock);
                goto SDPL_req_failed;
            }
            MOD_INC_WRITES(medium, minor);
        } else {
            MOD_INC_READS(medium, minor);
        }
#else
        if (rw == WRITE) {
            if (!(SDPL_map->bflags[SDPL_block] & SDPL_RW)) {
                sdpl_error("Try to write a read-only bloc !(" STD_ATTRIB
                           ")\n", SDPL_block);
                spin_unlock(minor_lock);
                goto SDPL_req_failed;
            }
            MOD_INC_WRITES(medium, minor);
        } else {
            MOD_INC_READS(medium, minor);
        }
#endif

        /* Adjust mapping for low level access */
        bh->b_rdev = dev_rdev;
        bh->b_rsector = dev_rsector;
        sdpl_debug("Remapped to sector %ld - device %04X (%u)\n",
                   dev_rsector, dev_rdev, SDPL_map->data_offset);
#if 0
        sdpl_info("Remapped to sector %ld - device %04X\n", dev_rsector,
                   dev_rdev);
#endif
        spin_unlock(minor_lock);
    } else {
        /* Request for a sdpl layer */
        if (minor < 0) {
            sdpl_error("Bad minor: %d\n", minor);
            spin_unlock(medium_lock);
            goto SDPL_req_failed;
        }
        bh->b_rdev = dev_rdev;
        spin_unlock(medium_lock);
    }
    return 1;

 SDPL_req_failed:
    buffer_IO_error(bh);
    return 0;
}

/** This function initialize internals & bloc device structures
 */
void SDPL_geninit (void)
{
    std_type medium, minor;

    sdpl_debug("%s -- SDPL_geninit\n", SDPL_name);

    blk_size[SDPL_MAJOR] = SDPL_sizes;
    blksize_size[SDPL_MAJOR] = SDPL_blocksizes;
    hardsect_size[SDPL_MAJOR] = SDPL_sectsizes;
    FOR_EACH_MEDIUM /* No semi-colum here, please !!! */
    {
        INIT_DEVS(medium);
        SDPL_LAYER(medium)->medium_lock = SPIN_LOCK_UNLOCKED;
        for (minor = 0; minor < MAX_MINOR; minor++) {
            SDPL_sizes[(medium << MEDIUM_SHIFT) + minor + 1] = 0;
#if 0
            SDPL_blocksizes[medium << MEDIUM_SHIFT + minor + 1] = BLOCK_SIZE;
            SDPL_sectsizes[medium << MEDIUM_SHIFT + minor + 1] = 512;
#else
            SDPL_blocksizes[(medium << MEDIUM_SHIFT) + minor + 1] = 0;
            SDPL_sectsizes[(medium << MEDIUM_SHIFT) + minor + 1] = 0;
#endif
            register_disk(NULL, MKDEV(SDPL_MAJOR,
                                      (medium << MEDIUM_SHIFT) + minor),
                          1, &SDPL_fops, 0);
        }
        SDPL_sizes[medium << MEDIUM_SHIFT] = 0;
#if 0
        SDPL_blocksizes[medium << MEDIUM_SHIFT] = BLOCK_SIZE;
        SDPL_sectsizes[medium << MEDIUM_SHIFT] = 512;
#else
        SDPL_blocksizes[medium << MEDIUM_SHIFT] = 0;
        SDPL_sectsizes[medium << MEDIUM_SHIFT] = 0;
#endif
    }
}

/** This function tries to free anything that stays allocated
 *   before throwing the driver away..
 */
void SDPL_freeall (void) {
    struct SDPL_map *SDPL_map = NULL;
    int medium, minor;

    sdpl_debug("SDPL_freeall\n");
    FOR_EACH_MEDIUM /* No semi-colum here, please !!! */
    {
        if (IS_DEV_ALLOC(medium))
        {
            if ((SDPL_map = SDPL_LAYER(medium)->SDPL_map) != NULL) {
                for (minor = 0; minor < SDPL_LAYER(medium)->max; minor++) {
                    free_partition(medium ,minor);
                }
            }
            free_device(SDPL_map, medium);
        }
    }
}

/** Driver initialization routine
 */
int SDPL_init (void)
{
    sdpl_debug("SDPL_init\n");

#ifdef MODULE
    if (sdpl_map && sdpl_map[0] != '\0') {
        sdpl_set_map(sdpl_map);
    }
#endif

    spin_lock(&SDPL_global_lock);
    /* Create devfs entries */
    if (REGISTER_DEVFS() < 0) {
        sdpl_error("Unable to get major number %d for " SDPL_DEVNAME
                   " device\n", SDPL_MAJOR);
        return -EIO;
    }
    /* Register bloc device */
    if (register_blkdev(SDPL_MAJOR, SDPL_name, &SDPL_fops)) {
        sdpl_error("Could not register blkdev %d", SDPL_MAJOR);
        return -EIO;
    }
    blk_queue_make_request(BLK_DEFAULT_QUEUE(SDPL_MAJOR), SDPL_make_request);

    /* Structures initialization */
    SDPL_geninit();

    /* Register into procfs */
    if (REGISTER_PROCFS() < 0) {
        sdpl_error("Unable to register proc interface\n");
        return -EFAULT;
    }
    spin_unlock(&SDPL_global_lock);
    sdpl_info("Module %s v%s successfully initialized\n",
              SDPL_name, SDPL_version);

    return 0;
}

/** Cleanup routine
 */
static void SDPL_cleanup(void)
{
    sdpl_debug("SDPL_cleanup\n");
    spin_lock(&SDPL_global_lock);

    /* Deregister devfs entries */
    CLEANUP_DEVFS();
    /* Deregister bloc device */
    if (unregister_blkdev(SDPL_MAJOR, SDPL_name) < 0) {
        sdpl_error("%s -- unregister_blkdev failed\n", SDPL_name);
    }
    blk_size[SDPL_MAJOR] = NULL;
    blksize_size[SDPL_MAJOR] = NULL;
    hardsect_size[SDPL_MAJOR] = NULL;

    /* Free all structures */
    SDPL_freeall();

    /* Deregister procfs entries */
    CLEANUP_PROCFS();
    sdpl_free_map();

    spin_unlock(&SDPL_global_lock);
    sdpl_debug(KERN_CRIT "Module %s successfully deactivated\n", SDPL_name);
    return;
}

/** Module registration
 */
module_init(SDPL_init);
module_exit(SDPL_cleanup);

/** Module description
 */
MODULE_AUTHOR("J. Mayer <l_indien at magic dot fr>");
MODULE_DESCRIPTION(SDPL_NAME);
MODULE_SUPPORTED_DEVICE(SDPL_DEVNAME);
#if (LINUX_VERSION_CODE >= 0x2040A)
MODULE_LICENSE("GPL");
#endif
