/* The goal of this module is to test the sdpl kernel ioctls */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/ioctl.h>

#include <linux/fs.h>
#include <linux/sdpl.h>

#define FLAG_NONE    0x00
#define FLAG_SILENT  0x01
#define FLAG_NOOPEN  0x02
#define FLAG_NOCLOSE 0x04

#define SDPL_BITS 8

static int getsize (const unsigned char *devname,
                    const unsigned char *line, const int flag);
static int flush (const unsigned char *devname,
                  const unsigned char *line, const int flag);
static int get_access (const unsigned char *devname,
                       const unsigned char *line, const int flag);
static int set_ro (const unsigned char *devname,
                   const unsigned char *line, const int flag);
static int set_rw (const unsigned char *devname,
                   const unsigned char *line, const int flag);
static int set_bootable (const unsigned char *devname,
                         const unsigned char *line, const int flag);
static int list_layers (const unsigned char *devname,
                        const unsigned char *line, const int flag);
static int get_bootable (const unsigned char *devname,
                         const unsigned char *line, const int flag);
#if 0
static int list_parts (const unsigned char *devname,
                       const unsigned char *line, const int flag);
static int get_bootpart (const unsigned char *devname,
                         const unsigned char *line, const int flag);
static int free_bloc (const unsigned char *devname,
                      const unsigned char *line, const int flag);
static int add_blocs (const unsigned char *devname,
                      const unsigned char *line, const int flag);
static int get_bl_access (const unsigned char *devname,
                          const unsigned char *line, const int flag);
static int set_bl_ro (const unsigned char *devname,
                      const unsigned char *line, const int flag);
static int set_bl_rw (const unsigned char *devname,
                      const unsigned char *line, const int flag);
#endif
static int dummy (const unsigned char *devname,
                  const unsigned char *line, const int flag);
static int help (const unsigned char *devname,
                 const unsigned char *line, const int flag);

struct command {
    const unsigned char *name;
    int (*action)(const unsigned char *devname,
                  const unsigned char *line, const int flag);
    const unsigned char *comments;
    char alternate;
};

const struct command cmds[] = {
    { "gets", &getsize, "Get the size of a device", '\0' },
    { "flush", &flush, "Flush device's caches", '\0' },
    { "setro", &set_ro, "Set a device read-only", '\0' },
    { "setrw", &set_rw, "Set a device read-write", '\0' },
    { "geta", &get_access, "Get a device's access mode", '\0' },
    { "setb", &set_bootable, "Set a device bootable", '\0' },
    { "getb", &get_bootable, "Get a device's bootable flag", '\0' },
    { "listl", &list_layers, "List all layers", '\0' },
#if 0
    { "listp", &list_parts, "List all sdpl partitions", '\0' },
    { "getbp", &get_bootpart, "Get the boot partition", '\0' },
    { "freeb", &free_bloc, "Deallocate a bloc from a partition", '\0' },
    { "addb", &add_blocs, "Add blocs to a partition", '\0' },
    { "getba", &get_bl_access, "Get access mode for a bloc", '\0' },
    { "setbro", &set_bl_ro, "Set a bloc read-only", '\0' },
    { "setrw", &set_bl_rw, "Set a bloc read_write", '\0' },
#endif
    { "dummy", &dummy, "Do nothing", '\0' },
    { "help", &help, "Get this help", '?' },
    { NULL, NULL, NULL, '\0' },
};

static int get_device (const unsigned char *name, const int flag)
{
    int fd;

    if (!(flag & FLAG_NOOPEN)) {
        if ((fd = open(name, O_RDONLY)) < 0) {
            fprintf(stderr, "Can't open device: %s\n", name);
            return -1;
        }
    } else {
        fd = 0;
    }

    return fd;
}

static int release_device (const int fd, const int flag)
{
    if (!(flag & FLAG_NOCLOSE)) {
        close(fd);
    }
    return 0;
}

#define print_result(flags, blabla, args...) \
    if (!(flag & FLAG_SILENT)) fprintf(stderr, blabla, ##args);

static int getsize (const unsigned char *devname,
                    const unsigned char *line, int flag)
{
    std_type size;
    int fd;

    if ((fd = get_device(devname, flag)) < 0)
        return -1;

    if (ioctl(fd, BLKGETSIZE, &size) < 0) {
        perror("Can't get size\n");
        return -1;
    }
    release_device(fd, flag);
    print_result(flag, "Size of device %s is " STD_ATTRIB " bytes\n",
                 devname, size);
    
    return (int)size;
}

static int flush (const unsigned char *devname,
                  const unsigned char *line, int flag)
{
    int fd;

    if ((fd = get_device(devname, flag)) < 0)
        return -1;

    if (ioctl(fd, BLKFLSBUF, NULL) < 0) {
        perror("Can't flush device\n");
        return -1;
    }
    release_device(fd, flag);
    print_result(flag, "Device %s flushed\n", devname);
    
    return 0;
}

static int get_access (const unsigned char *devname,
                       const unsigned char *line, int flag)
{
    int access_mode;
    int fd;

    if ((fd = get_device(devname, flag)) < 0)
        return -1;

    if (ioctl(fd, SDPL_GET_ACCESS, &access_mode) < 0) {
        perror("Can't get access mode\n");
        return -1;
    }
    release_device(fd, flag);
    print_result(flag, "Device %s access mode is read-%s\n", devname,
                 (access_mode == SDPL_RO) ? "only" : "write");
    return access_mode;
}

static int set_ro (const unsigned char *devname,
                   const unsigned char *line, int flag)
{
    int fd;
    int access_mode;

    if ((access_mode = get_access(devname, NULL, flag | FLAG_SILENT)) < 0)
        return -1;
        
    if (access_mode == SDPL_RO) {
        print_result(flag, "Device %s is already read-only\n", devname);
        return 0;
    }
    if ((fd = get_device(devname, flag)) < 0)
        return -1;
    access_mode = SDPL_RO;
    if (ioctl(fd, SDPL_SET_ACCESS, &access_mode) < 0) {
        perror("Can't set device read-only");
        return -1;
    }
    return access_mode;
}

static int set_rw (const unsigned char *devname,
                   const unsigned char *line, int flag)
{
    int fd;
    int access_mode;

    if ((access_mode = get_access(devname, NULL, flag | FLAG_SILENT)) < 0)
        return -1;
        
    if (access_mode == SDPL_RW) {
        print_result(flag, "Device %s is already read-write\n", devname);
        return 0;
    }
    if ((fd = get_device(devname, flag)) < 0)
        return -1;
    access_mode = SDPL_RW;
    if (ioctl(fd, SDPL_SET_ACCESS, &access_mode) < 0) {
        perror("Can't set device read-only");
        return -1;
    }
    return access_mode;
}

static int get_bootable (const unsigned char *devname,
                         const unsigned char *line, int flag)
{
    int bootable;
    int fd;

    if ((fd = get_device(devname, flag)) < 0)
        return -1;

    if (ioctl(fd, SDPL_GET_BOOTABLE, &bootable) < 0) {
        perror("Can't get boot flag\n");
        return -1;
    }
    release_device(fd, flag);
    print_result(flag, "Device %s access mode is %s bootable\n", devname,
                 bootable ? "" : "not");
    return bootable;
}

static int set_bootable (const unsigned char *devname,
                         const unsigned char *line, int flag)
{
    int fd;
    int bootable;

    if ((bootable = get_access(devname, NULL, flag | FLAG_SILENT)) < 0)
        return -1;
        
    if (bootable) {
        print_result(flag, "Device %s is already bootable\n", devname);
        return 0;
    }
    if ((fd = get_device(devname, flag)) < 0)
        return -1;
    if (ioctl(fd, SDPL_SET_BOOTABLE, NULL) < 0) {
        perror("Can't mark device bootable");
        return -1;
    }
    print_result(flag, "Device %s set bootable\n", devname);
    return 1;
}

static unsigned char *find_eol (const unsigned char *line, const int count)
{
    unsigned char *pos = (unsigned char *)line;
    
    while (pos - line < count) {
        if (*pos == '\n')
            return pos;
        pos++;
    }
    return NULL;
}

static int get_number (const unsigned char *string)
{
    return atoi(string);
}

static int list_layers (const unsigned char *devname,
                        const unsigned char *line, const int flag)
{
    unsigned char gline[128];
    unsigned char procname[32];
    unsigned char *pos = gline, *next;
    int got, remain = 0;
    int num_parts = 0;
    int fd;

    snprintf(procname, 32, "/proc/%s/layers", SDPL_DEVNAME);
    if ((fd = get_access(procname, NULL, flag)) < 0) {
        print_result(flag, "Unable to open %s\n", procname);
        return -1;
    }
    while ((got = read(fd, pos, sizeof(gline) - remain)) > 0 && remain > 0) {
        remain += got;
        while ((next = find_eol(pos, remain)) != NULL) {
            *next = '\0';
            next++;
            /* The only interresting line is SDPL map... */
            if (memcmp(pos, "Num parts", 9) == 0) {
                num_parts += get_number(pos + 10);
                pos = next;
                continue;
            }
            if (memcmp(pos, "SDPL map", 8) == 0) {
                num_parts += 1 << SDPL_BITS;
                pos = next;
                pos = next;
            }
            /* All other lines are skipped */
            pos = next;
        }
        remain = gline + sizeof(gline) - pos;
        memcpy(gline, pos, remain);
        pos = gline;
    }
    print_result(flag, "Found %d layers - %d partitions\n", 
                 num_parts >> SDPL_BITS, num_parts & ((1 << SDPL_BITS) - 1));

    return num_parts;
}

static int help (const unsigned char *devname,
                 const unsigned char *line, const int flag)
{
    const struct command *command = cmds;

    while (command->name != NULL) {
        print_result(flag & ~FLAG_SILENT"%s: %s\t%s\n", command->name, 
                     command->alternate != '\0', command->comments);
        command++;
    }

    return 0;
}

static int dummy (const unsigned char *devname,
                  const unsigned char *line, const int flag)
{
    return 0;
}

#define CMDLEN 256

static int find_not_space (const unsigned char *buffer, unsigned char **pos)
{
    unsigned char *search = (unsigned char *)buffer;
    int len = strlen(buffer);

    while ((search - buffer) < len) {
        if (!isspace(*search)) {
            *pos = search;
            return (search - buffer);
        }
        search++;
    }
    *pos = NULL;
    return -1;
}

int main (void)
{
    const struct command *command;
    unsigned char cmd_buf[CMDLEN];
    unsigned char *first, *pos, *pos2;
    int cmd_len;
    int i;

    while (1) {
    do_loop:
        fgets(cmd_buf, CMDLEN, stdin);

        cmd_len = strlen(cmd_buf);
        i = find_not_space(cmd_buf, &first);
        /* All spaces -> next one */
        if (i == -1) {
            dummy(NULL, NULL, 0);
            continue;
        }

        pos = first;
        cmd_len = strlen(first);
        while ((pos - cmd_buf) < cmd_len) {
            if (isspace(*pos)) {
                *pos = '\0';
                break;
            }
            pos++;
        }
        if ((pos - cmd_buf) == cmd_len) {
            help(NULL, NULL, 0);
            continue;
        }
        pos++;
        pos2 = first;
        while (pos2 < pos) {
            *pos2 = tolower(*pos2);
            pos2++;
        }

        command = cmds;
        while (command->name != NULL) {
            if (strcmp(first, command->name) == 0) {
                pos2 = pos;
                cmd_len = strlen(pos);
                while ((pos2 - pos) < cmd_len) {
                    if (isspace(*pos2)) {
                        *pos2 = '\0';
                        break;
                    }
                    pos2++;
                }
                command->action(pos, pos2, 0);
                goto do_loop;
            }
            command++;
        }
        /* Bad command */
        if (command->name == NULL) {
            sdpl_error("Error: cmd not found (%c)\n", *pos);
            help(NULL, NULL, 0);
            continue;
        }
    }

    return 0;
}
