/*
 * SDPL sdplmount.c: a mount utility to tell the kernel to parse a layer
 *
 *   Copyright (C) 2001, 2002, 2003, 2004, Jocelyn MAYER <l_indien@magic.fr>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
/*
 * Simple Disk Partition Layer
 *
 * Utility to register / unregister a sdpl layer into the kernel
 *
 */

/**SECTION: sysutils */
/**SECTION: sdpl */
/**SECTION: mount tool */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
/* This is the GLIBC definition to use 64 bits file support */
#if CONFIG_BLK_SDPL_64BITS
#ifndef __USE_LARGEFILE64
#define __USE_LARGEFILE64 1
#endif
#ifndef __USE_FILE_OFFSET64
#define __USE_FILE_OFFSET64 1
#endif
#else /* CONFIG_BLK_SDPL_64BITS */
#ifndef __USE_LARGEFILE64
#define __USE_LARGEFILE64 0
#endif
#ifndef __USE_FILE_OFFSET64
#define __USE_FILE_OFFSET64 0
#endif
#endif /* CONFIG_BLK_SDPL_64BITS */
#include <fcntl.h>
#include <asm/byteorder.h>
#include <linux/types.h>
#include <linux/limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/ioctl.h>

#include <linux/sdpl.h>

static const char *sdplm_error[] = {
    "Bad number of arguments",
    "Can't open sdpl device ",
    "Can't open device ",
    "Error while releasing layer ",
    "Error while parsing layer ",
};
enum { SDPLM_BADARGS = 0, SDPLM_NOSDPL, SDPLM_NODEV,
       SDPLM_UMOUNT, SDPLM_MOUNT };

/** Usage...
 */
static void usage (void)
{
    fprintf(stdout, "sdplmount usage: mount / unmount a bloc device "
            " as a sdpl layer\n");
    fprintf(stdout, "sdplmount [-d] [?] [-h] device sdpl_dev\n");
    fprintf(stdout, " -d: unmount\n");
    fprintf(stdout, " -? -h: view this help\n");
}

/** Main entry point
 */
int main (int argc, char **argv)
{
    char c;
    int umount = 0;
    int sdpl_fd, blk_fd;
    int ret = -1;
    char *sdpl_name, *blk_name;

    /* Get options */
    while ((c = getopt(argc, argv, "dh?")) != (char)EOF) {
        switch (c) {
        case 'd':
            umount = 1;
            break;
        case '?':
        case 'h':
            usage();
            return 0;
        }
    }
    if (argc != 3) {
        fprintf(stderr, "%s\n", sdplm_error[SDPLM_BADARGS]);
        usage();
        return -1;
    }

    sdpl_name = argv[2];
    sdpl_debug("About to open sdpl layer: %s\n", sdpl_name);
#if CONFIG_BLK_SDPL_EMBEDDED
    if ((sdpl_fd = open(sdpl_name, O_RDWR)) < 0) {
        perror(sdplm_error[SDPLM_NOSDPL]);
        return -1;
    }
#else /* CONFIG_BLK_SDPL_EMBEDDED */
    if ((sdpl_fd = open(sdpl_name, O_RDWR | O_LARGEFILE)) < 0) {
        perror(sdplm_error[SDPLM_NOSDPL]);
        return -1;
    }
#endif /* CONFIG_BLK_SDPL_EMBEDDED */
    
    if (umount == 1) {
        blk_name = NULL;
        if (ioctl(sdpl_fd, SDPL_DETACH, NULL) < 0) {
            fprintf(stderr, "%s %s\n", sdplm_error[SDPLM_UMOUNT], sdpl_name);
            perror("");
            goto mount_end;
        } else {
            ret = 0;
        }
    } else {
        blk_name = argv[1];
        sdpl_debug("About to parse device: %s\n", blk_name);
#if CONFIG_BLK_SDPL_EMBEDDED
        if ((blk_fd = open(blk_name, O_RDWR)) < 0) {
            perror(sdplm_error[SDPLM_NODEV]);
            goto mount_end;
        }
#else /* CONFIG_BLK_SDPL_EMBEDDED */
        if ((blk_fd = open(blk_name, O_RDWR | O_LARGEFILE)) < 0) {
            perror(sdplm_error[SDPLM_NODEV] );
            goto mount_end;
        }
#endif /* CONFIG_BLK_SDPL_EMBEDDED */
        if (ioctl(sdpl_fd, SDPL_ATTACH, blk_fd) < 0) {
            fprintf(stderr, "%s %s\n", sdplm_error[SDPLM_MOUNT], sdpl_name);
            perror("");
            ret = -1;
        } else {
            ret = 0;
        }
        close(blk_fd);
    }
 mount_end:
    close(sdpl_fd);

    return ret;
}

