/*
 * bd_remap.c
 *
 * This module intend to remap a bloc device into memory.
 * The application of this device is to manipulate
 * the root partition while the system is running,
 * with no risk to crash if the kernel or an application
 * intend to access to a file on the device.
 */
/*
 * Copyright (C) 2003-2004 by Jocelyn Mayer <l_indien@magic.fr>
 * distributed under the GNU Public License
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
/* History:
 * 2003-01-25: Created
 */

/* Theory of operations:
 * This device works like a ramdisk.
 * It will be "attached" to another block device,
 * given it's (major, minor).
 * It can also be given a list of blocs to treat in a special way.
 * This can be: - use the original device.
 *              - send back an error, as if the bloc is invalid.
 * It will then allocate the needed space in the cache,
 * quite like a ramdisk does.
 * To save some space, we won't copy empty blocs,
 * but use a special one which will be remapped only
 * if someone tries to write in it, with no-zero data.
 * To be smooth with  flash devices, we will also treat blocs
 * filled with 0xFF the same way.
 */

#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/unistd.h>
#include <linux/slab.h>
#include <linux/hdreg.h>
#include <linux/file.h>
#include <asm/uaccess.h>
#include <linux/blk.h>
#include <linux/ioctl.h>
#include <linux/vmalloc.h>

//#include <linux/bd_remap.h>
#include "bd_remap.h"

//#define DEBUG_REMAP
#define DEBUG_INFO
#ifdef DEBUG_REMAP
#undef  REMAP_DEBUG
#define REMAP_DEBUG(fmt, args...) do { printk(fmt , ##args); } while (0)
#else
#define REMAP_DEBUG(fmt, args...)
#endif
#undef  REMAP_ERROR
#define REMAP_ERROR(fmt, args...) do { printk(fmt , ##args); } while (0)

/* Devfs stuff */
#if LINUX_VERSION_CODE >= 0x020300 && !defined(USE_NO_DEVFS)
static devfs_handle_t devfs_handle;
#endif

#if defined(USE_BDREMAP_MUTEXES)
/* By hand mutexes */
#define __mutex_INITIALIZER(name)      			                \
{									\
	.count		= ATOMIC_INIT(1),				\
	.wait		= __WAIT_QUEUE_HEAD_INITIALIZER((name).wait)	\
}

#define DECLARE_mutex(name) \
	mutex_t name = __mutex_INITIALIZER(name)

static inline void init_mutex (struct mutex *mtx)
{
    atomic_set(&mtx->count, 1);
    init_waitqueue_head(&mtx->wait);
}

/*
 * Atomically update mtx->count.
 */
static inline int __mutex_update_count (struct mutex *mtx, int incr)
{
    int old_count, tmp;
    static spinlock_t mutexes_lock = SPIN_LOCK_UNLOCKED;
    unsigned long flags;

    spin_lock_irqsave(&mutexes_lock, flags);
    old_count = atomic_read(&mtx->count);
#if 0
    tmp = max_t(int, old_count, 0) + incr;
#else
    tmp = ((~(old_count >> 31)) & old_count) + incr;
#endif
    atomic_set(&mtx->count, tmp);
    spin_unlock_irqrestore(&mutexes_lock, flags);
    
    return old_count;
}

/*
 * Note that when we come in to __mutex_lock,
 * we have already decremented count, but that decrement was
 * ineffective since the result was < 0, and any negative value
 * of count is equivalent to 0.
 * Thus it is only when we decrement count from some value > 0
 * that we have actually got the mutex.
 */
static void __mutex_lock (struct mutex *mtx)
{
	struct task_struct *tsk = current;
	DECLARE_WAITQUEUE(wait, tsk);

	__set_task_state(tsk, TASK_UNINTERRUPTIBLE);
	add_wait_queue_exclusive(&mtx->wait, &wait);

	/*
	 * Try to get the mutex.  If the count is > 0, then we've
	 * got the mutex; we decrement count and exit the loop.
	 * If the count is 0 or negative, we set it to -1, indicating
	 * that we are asleep, and then sleep.
	 */
	while (__mutex_update_count(mtx, -1) <= 0) {
		schedule();
		set_task_state(tsk, TASK_UNINTERRUPTIBLE);
	}
	remove_wait_queue(&mtx->wait, &wait);
	__set_task_state(tsk, TASK_RUNNING);

	/*
	 * If there are any more sleepers, wake one of them up so
	 * that it can either get the mutex, or set count to -1
	 * indicating that there are still processes sleeping.
	 */
	wake_up(&mtx->wait);
}

static void __mutex_unlock (struct mutex *mtx)
{
	/*
	 * Note that we incremented count in mutex_unlock() before we came here,
	 * but that was ineffective since the result was <= 0, and
	 * any negative value of count is equivalent to 0.
	 * This ends up setting count to 1, unless count is now > 0
	 * (i.e. because some other cpu has called up() in the meantime),
	 * in which case we just increment count.
	 */
	__mutex_update_count(mtx, 1);
	wake_up(&mtx->wait);
}

static inline void mutex_lock (struct mutex *mtx)
{
    /*
     * Try to get the mutex, take the slow path if we fail.
     */
    if (unlikely(atomic_dec_return(&mtx->count) < 0))
        __mutex_lock(mtx);
}

static inline void mutex_unlock (struct mutex *mtx)
{
    if (unlikely(atomic_inc_return(&mtx->count) <= 0))
        __mutex_unlock(mtx);
}
#else
#define DECLARE_mutex(m) DECLARE_MUTEX(m)
#define init_mutex(m)    init_MUTEX(m)
#define mutex_lock(m)    down(m)
#define mutex_unlock(m)  up(m)
#endif

/* Data blocs are stored in this list */
static unsigned char **data_blocs;
static unsigned long nb_used, nb_allocs;
static DECLARE_mutex(data_lock);

/* Remapped devices are described here */
static rem_dev_t *dev_list;
static unsigned short nb_remaps;
static DECLARE_mutex(devices_lock);

#define GET_MAJOR(dev) ((dev) >> 8)
#define GET_MINOR(dev) ((dev) & 0xFF)
#define MAKEDEV(major, minor) (((major) << 8) | (minor))

/* Helpers for mutex usage */
static inline void lock_data (void)
{
    mutex_lock(&data_lock);
}

static inline void unlock_data (void)
{
    mutex_unlock(&data_lock);
}

static inline void lock_devices (void)
{
    mutex_lock(&devices_lock);
}

static inline void unlock_devices (void)
{
    mutex_unlock(&devices_lock);
}

static inline void lock_device (int dev)
{
    mutex_lock(&dev_list[dev].lock);
}

static inline void unlock_device (int dev)
{
    mutex_unlock(&dev_list[dev].lock);
}

static inline void lock_all (void)
{
    lock_devices();
    lock_data();
}

static inline void unlock_all (void)
{
    unlock_data();
    unlock_devices();
}

/* This routine makes direct access to the underlaying bloc devices
 * through our new device.
 */
static int remap_bloc_request (request_queue_t *q, int rw,
                               struct buffer_head *bh)
{
    rem_dev_t *device;
    unsigned short minor;
    
    minor = GET_MINOR(bh->b_rdev);
    /* dev_list should only be NULL if nb_remaps == 0, but... */
    if (minor >= nb_remaps || dev_list == NULL) {
        REMAP_ERROR("ERROR: minor: %d nb_remaps: %d dev_list:%p\n",
                    minor, nb_remaps, dev_list);
        buffer_IO_error(bh);
        return 0;
    }
    device = &dev_list[minor];
    REMAP_DEBUG("%s: Read sect %lu to phy dev (%d->[%x %x] - %d)\n",
                __func__, bh->b_rsector,
                GET_MAJOR(bh->b_rdev), device->device >> 8,
                device->device & 0xFF, minor);
    bh->b_rdev = device->device;

    return (*device->default_request)(q, rw, bh);
}

/* We handle all access requests here.
 */
static int remap_make_request (request_queue_t *q, int rw, 
                               struct buffer_head *bh)
{
    rem_dev_t *device;
    unsigned char *buffer;
    kdev_t dev;
    unsigned long int bloc, rbloc;
    int len, tmp_len;
    unsigned int offset;
    unsigned short major, minor;
    int i;
    unsigned char pattern;

    major = GET_MAJOR(bh->b_rdev);
    minor = GET_MINOR(bh->b_rdev);
    REMAP_DEBUG("Access to device: (%d-%d %x)\n", major, minor, bh->b_rdev);

#ifdef READA
    if (rw == READA)
        rw = READ;
#endif

    /* Find the remapped device in our list */
    device = NULL;
    dev = (kdev_t)(-1);
    for (i = 0; i < nb_remaps; i++) {
        device = &dev_list[i];
        if (GET_MAJOR(device->device) == major)
            dev = device->device;
        if (device->device == bh->b_rdev)
            break;
    }
    /* If we don't know about this device,
     * that means we remapped another device with the same major.
     * If we have no blocs described for this device, we do the same thing.
     */
    if (i == nb_remaps || device->btype == NULL) {
        if (dev == (kdev_t)(-1)) {
            REMAP_ERROR("No device for access (%d - %d)\n", major, minor);
            goto remap_request_failed;
        }
        REMAP_DEBUG("Direct access to: (%d-%d) through (%d-%d) (%lu)\n",
                    GET_MAJOR(dev), GET_MINOR(dev), major, minor,
                    bh->b_rsector);
        return (*device->default_request)(q, rw, bh);
    }

    len = bh->b_size;
    offset = (bh->b_rsector & ((1 << device->bloc_dec ) - 1)) << 9;
    bloc = bh->b_rsector >> device->bloc_dec;
    buffer = bh_kmap(bh);
    REMAP_DEBUG("Remap: %lu - %lu (%d - %d %d %d)\n", bh->b_rsector,
                bloc, len, offset, device->bloc_dec, i);
    if (bloc >= device->nb_blocs) {
        REMAP_ERROR("try to access beyond end of device\n");
        goto remap_request_failed;
    }
    /* As our bloc size can be different from the bloc size of the 
     * request, we need to loop here.
     */
    while (len > 0) {
        rbloc = device->btype[bloc];
        tmp_len = (REMAP_BLOC_SIZE << device->bloc_dec) - offset;
        if (tmp_len > len)
            tmp_len = len;
        switch (rbloc) {
        case REMAP_DISCARD:
            /* Invalid request: bloc hasn't been remapped in any way */
            REMAP_DEBUG("Bloc %lu isn't remapped...\n", bloc);
            goto remap_request_failed;

        case REMAP_NOREMAP:
#if 0 // Not supported by now: Need to generate a new request...
            /* Remap to the original device */
            bh->b_rdev = device->device;
            return (*device->default_request)(q, rw, bh);
#else
            REMAP_DEBUG("Bloc %lu is on the original device...\n", bloc);
            goto remap_request_failed;
#endif
            
        case REMAP_EMPTY:
            REMAP_DEBUG("Bloc %lu is empty...\n", bloc);
            pattern = 0x00;
            if (rw == WRITE)
                goto remap_request_failed;
            goto fill_bloc;

        case REMAP_EMPTYFF:
            REMAP_DEBUG("Bloc %lu is filled with 0xFF...\n", bloc);
            pattern = 0xFF;
            if (rw == WRITE)
                goto remap_request_failed;
            goto fill_bloc;
            
        case 0:
            REMAP_DEBUG("Access to bloc 0...\n");
            pattern = 0x00;
            if (rw == WRITE)
                goto copy_on_write;
            goto fill_bloc;

        case 1:
            REMAP_DEBUG("Access to bloc 1...\n");
            pattern = 0xFF;
            if (rw == WRITE)
                goto copy_on_write;
            goto fill_bloc;

        default:
            rbloc -= 2;
            /* Remap to device's copy... */
            goto serv_bloc;
        }
    fill_bloc:
        memset(buffer, pattern, tmp_len);
        goto end_bloc;

    copy_on_write:
        /* If we tries to write in bloc 0 or 1,
         * we do a copy of the buffer before serving the write request
         */
        REMAP_DEBUG("Bloc %lu : copy on write...\n", bloc);
        /* No need to reallocate data_blocs here... */
        data_blocs[nb_used] = vmalloc(REMAP_BLOC_SIZE << device->bloc_dec);
        if (data_blocs[nb_used] == NULL) {
            REMAP_ERROR("Can't reallocate bloc for copy-on-write req\n");
            goto remap_request_failed;
        }
        memset(data_blocs[nb_used], pattern,
               REMAP_BLOC_SIZE << device->bloc_dec);
        device->btype[bloc] = REMAP_COPY;
        device->btype[bloc] |= nb_used + 2;
        device->nb_mapped++;
        rbloc = nb_used++;
        
    serv_bloc:
        if (rbloc >= device->nb_mapped) {
            REMAP_ERROR("try to access beyond end of device\n");
            goto remap_request_failed;
        }
        REMAP_DEBUG("Remap: %lu (%lu - %d) to %d (%d)\n",
                    bh->b_rsector, bloc, offset,
                    device->btype[bloc], tmp_len);
        if (rw == WRITE)
            memcpy(data_blocs[rbloc] + offset, buffer, tmp_len);
        else
            memcpy(buffer, data_blocs[rbloc] + offset, tmp_len);
                   
    end_bloc:
        buffer += tmp_len;
        len -= tmp_len;
        bloc++;
        offset = 0;
    }

    if (rw == READ)
        flush_dcache_page(bh->b_page);
    bh_kunmap(bh);
    bh->b_end_io(bh, 1);

    return 0;

 remap_request_failed:
    buffer_IO_error(bh);

    return 0;
}

/* Bloc devices operations */
static int remap_blk_open(struct inode *inode, struct file *file)
{
    unsigned short minor;

    REMAP_DEBUG("blk_open %x\n", inode->i_rdev);
    minor = GET_MINOR(inode->i_rdev);
    if (minor >= nb_remaps || dev_list[minor].btype == NULL) {
        REMAP_ERROR("invalid open\n");
        return -ENODEV;
    }
    MOD_INC_USE_COUNT;

    return 0;
}

static int remap_blk_close(struct inode *inode, struct file *file)
{
    REMAP_DEBUG("blk_close %x\n", inode->i_rdev);
    MOD_DEC_USE_COUNT;

    return 0;
}

/* Control char device operations */
static int remap_chr_open(struct inode *inode, struct file *file)
{
    unsigned short minor;

    REMAP_DEBUG("chr_open %x\n", inode->i_rdev);
    minor = GET_MINOR(inode->i_rdev);
    if (minor > 0) {
        REMAP_ERROR("invalid open on char device\n");
        return -ENODEV;
    }
    MOD_INC_USE_COUNT;

    return 0;
}

static int remap_chr_close(struct inode *inode, struct file *file)
{
    REMAP_DEBUG("chr_close %x\n", inode->i_rdev);
    MOD_DEC_USE_COUNT;

    return 0;
}

static int remap_domap (unsigned char *addr)
{
    user_remap_t user_remap;
    struct buffer_head *bh;
    rem_dev_t *new_dev_list, *device;
    unsigned char **new_data_blocs, *cur_addr, *buffer, *tmp_buffer, *pos;
    unsigned int *ipos;
    kdev_t rdev;
    unsigned int len, tmp_len;
    int all_zero, all_ff;
    int i, j, k;
    int ret;

    lock_all();
    copy_from_user(&user_remap, addr, sizeof(user_remap_t));
    rdev = user_remap.dev;
    REMAP_DEBUG("Was asked to map %lu blocs (%d) from device (%d-%d)\n",
                user_remap.nb_blocs, user_remap.bloc_dec,
                GET_MAJOR(rdev), GET_MINOR(rdev));
    /* Check that we didn't already remapped this device */
    for (i = 0; dev_list != NULL && i < nb_remaps; i++) {
        device = &dev_list[i];
        if (dev_list[i].device == rdev) {
            REMAP_ERROR("Device already mapped\n");
            unlock_all();
            return -EINVAL;
        }
    }

    new_data_blocs = NULL;
    device = NULL;
    tmp_buffer = NULL;
    /* Real work start here... */
    /* realloc devices list */
    REMAP_DEBUG("Realloc device list\n");
    new_dev_list = kmalloc((nb_remaps + 1) * sizeof(rem_dev_t), GFP_KERNEL);
    if (new_dev_list == NULL) {
        ret = -ENOMEM;
        REMAP_ERROR("Unable to allocate new device list\n");
        goto map_error;
    }
    memcpy(new_dev_list, dev_list, nb_remaps * sizeof(rem_dev_t));
    /* Realloc data blocs list */
    REMAP_DEBUG("Realloc data blocs from %lu to %lu\n", nb_used,
                nb_used + user_remap.nb_blocs);
    new_data_blocs =
        vmalloc((nb_used + user_remap.nb_blocs) * sizeof(char *));
    if (new_data_blocs == NULL) {
        ret = -ENOMEM;
        REMAP_ERROR("Unable to allocate new blocs list\n");
        goto map_error;
    }
    memcpy(new_data_blocs, data_blocs, sizeof(char *) * nb_used);
    memset(new_data_blocs + nb_used, 0,
           user_remap.nb_blocs * sizeof(char *));
    
    /* Initialize new device's structure */
    REMAP_DEBUG("Init device\n");
    device = &new_dev_list[nb_remaps];
    memset(device, 0, sizeof(rem_dev_t));
    init_mutex(&device->lock);
    fsync_dev(rdev);
    device->device = rdev;
    device->bloc_dec = user_remap.bloc_dec;
    device->default_request = (blk_get_queue(rdev))->make_request_fn;

    REMAP_DEBUG("allocate blocs description table\n");
    device->btype = vmalloc(user_remap.nb_blocs * sizeof(int));
    if (device->btype == NULL) {
        REMAP_ERROR("Unable to allocate blocs description table\n");
        ret = -ENOMEM;
        goto map_error;
    }
    /* Copy attributes for blocs */
    cur_addr = addr + sizeof(user_remap_t);
    copy_from_user(device->btype, cur_addr, sizeof(int) * user_remap.nb_blocs);
    tmp_buffer = vmalloc(REMAP_BLOC_SIZE << device->bloc_dec);
    if (tmp_buffer == NULL) {
        REMAP_ERROR("Unable to allocate a temporary buffer\n");
        ret = -ENOMEM;
        goto map_error;
    }

    for (i = 0, j = 0; i < user_remap.nb_blocs; i++) {
        switch (device->btype[i]) {
        case REMAP_EMPTY:
        case REMAP_EMPTYFF:
        case REMAP_DISCARD:
        case REMAP_NOREMAP:
            break;
        default:
            device->btype[i] = REMAP_COPY;
        case REMAP_COPY:
            /* Read bloc from the original device */
            REMAP_DEBUG("Read bloc %lx, %d\n", (unsigned long)rdev, i);
            len = REMAP_BLOC_SIZE << device->bloc_dec;
            pos = tmp_buffer;
            while (len > 0) {
                tmp_len = len;
                if (tmp_len > REMAP_BLOC_SIZE)
                    tmp_len = REMAP_BLOC_SIZE;
                bh = bread(rdev, i, tmp_len);
                if (bh == NULL) {
                    ret = -EFAULT;
                    REMAP_ERROR("Unable to read bloc %d\n", i);
                    goto map_error;
                }
                /* Test if the bloc is empty or all 0xFF */
                buffer = bh_kmap(bh);
                memcpy(pos, buffer, tmp_len);
                bh_kunmap(bh);
                brelse(bh);
                pos += tmp_len;
                len -= tmp_len;
            }
            all_zero = 1;
            all_ff = 1;
            ipos = (unsigned int *)tmp_buffer;
            for (k = 0;
                 k < ((REMAP_BLOC_SIZE << device->bloc_dec) >> 2); k++) {
                if (*ipos != 0)
                    all_zero = 0;
                if (*ipos != (unsigned int)(-1))
                    all_ff = 0;
                if (all_ff == 0 && all_zero == 0)
                    break;
                ipos++;
            }
            if (all_zero == 1) {
                REMAP_DEBUG("Bloc is all zeroes !\n");
                device->btype[i] |= 0;
                break;
            }
            if (all_ff == 1) {
                REMAP_DEBUG("Bloc is all 0xFF !\n");
                device->btype[i] |= 1;
                break;
            }
            REMAP_DEBUG("allocate bloc %d (%lu)\n", j, j + nb_used);
            device->btype[i] |= j + nb_used + 2;
            new_data_blocs[j + nb_used] =
                vmalloc(REMAP_BLOC_SIZE << device->bloc_dec);
            if (new_data_blocs[j + nb_used] == NULL) {
                REMAP_ERROR("Unable to allocate bloc %d\n", j);
                ret = -ENOMEM;
                goto map_error;
            }
            memcpy(new_data_blocs[j + nb_used],
                   tmp_buffer, REMAP_BLOC_SIZE << device->bloc_dec);
            j++;
            break;
        }
    }
    blk_queue_make_request(BLK_DEFAULT_QUEUE(GET_MAJOR(rdev)),
                           remap_make_request);
    
    /* We should reallocated the bloc table to its real size:
     * empty blocs have been allocated and not used...
     */
    if (tmp_buffer != NULL)
        vfree(tmp_buffer);
    vfree(data_blocs);
    data_blocs = new_data_blocs;
    kfree(dev_list);
    dev_list = new_dev_list;

    device->nb_blocs = user_remap.nb_blocs;
    nb_allocs += user_remap.nb_blocs;
    device->nb_mapped = j;
    nb_used += j;

    /* Send back to the user characteristics of the remapped device */
    REMAP_DEBUG("send data back to userland (%lu - %lu -> %lu - %lu)\n",
                device->nb_blocs, device->nb_mapped, nb_allocs, nb_used);
    memset(&user_remap, 0, sizeof(user_remap_t));
    user_remap.dev = MAKEDEV(REMAP_MAJOR, nb_remaps++);
    unlock_all();
    copy_to_user(addr, &user_remap, sizeof(user_remap_t));
    
    return 0;

map_error:
    if (tmp_buffer != NULL)
        vfree(tmp_buffer);
    for (j = 0; j < user_remap.nb_blocs; j++) {
        if (new_data_blocs[j + nb_used] != NULL)
            vfree(new_data_blocs[j + nb_used]);
    }
    if (device != NULL) {
        vfree(device->btype);
        device->btype = NULL;
    }
    vfree(new_data_blocs);
    kfree(new_dev_list);
    unlock_all();

    return ret;
}

static int remap_dounmap (unsigned char *addr)
{
    user_remap_t user_remap;
    rem_dev_t *new_dev_list, *device;
    unsigned char **new_data_blocs;
    kdev_t rdev;
    unsigned long numbloc;
    unsigned short major;
    int i, j, k, nb;
    int ret;

    lock_all();
    copy_from_user(&user_remap, addr, sizeof(user_remap_t));
    rdev = user_remap.dev;
    REMAP_DEBUG("Unmap: %x\n", user_remap.dev);
    major = GET_MAJOR(rdev);
    /* Check the device is really mapped */
    for (i = 0; i < nb_remaps; i++) {
        if (dev_list[i].device == rdev) {
            /* Restore make_request function if there is no device
             * with the same major mapped
             */
            device = &dev_list[i];
            fsync_dev(device->device);
            __invalidate_buffers(device->device, 1);
            nb = 0;
            for (j = 0; j < nb_remaps; j++) {
                if (GET_MAJOR(dev_list[i].device) == major)
                    nb++;
            }
            new_data_blocs = NULL;
            if (nb == 1) {
                /* Restore make_request handler */
                blk_queue_make_request(BLK_DEFAULT_QUEUE(major),
                                       device->default_request);
            }
            /* realloc devices list */
            new_dev_list = kmalloc((nb_remaps - 1) * sizeof(rem_dev_t),
                                   GFP_KERNEL);
            if (new_dev_list == NULL) {
                ret = -ENOMEM;
                REMAP_ERROR("can't realloc device list\n");
                goto unmap_error;
            }
            for (j = 0, k = 0; j < nb_remaps; j++) {
                if (j != i) {
                    memcpy(&new_dev_list[k++], &dev_list[j],
                           sizeof(rem_dev_t));
                }
            }
            REMAP_DEBUG("Set blocs from %lu to %lu\n", nb_allocs,
                   nb_allocs - device->nb_blocs);
            /* Realloc data blocs list */
            if (nb_allocs > device->nb_blocs) {
                new_data_blocs =
                    vmalloc((nb_allocs - device->nb_blocs) * sizeof(char *));
                if (new_data_blocs == NULL) {
                    ret = -ENOMEM;
                    REMAP_ERROR("can't realloc data blocs\n");
                    goto unmap_error;
                }
                /* mark all blocs from the device invalids */
                for (j = 0, k = 0; j < device->nb_blocs; j++) {
                    /* Safety checks (more than needed) */
                    if (device->btype[j] >= 1) {
                        numbloc = device->btype[j] - 2;
                        vfree(data_blocs[numbloc]);
                        data_blocs[numbloc] = NULL;
                        k++;
                    }
                }
                /* Safety check */
                if (k != device->nb_mapped) {
                    ret = -EFAULT;
                    REMAP_ERROR("mapped bloc count differ: %d %lu\n",
                                k, device->nb_mapped);
                    goto unmap_error;
                }
                /* Copy all valid blocs */
                for (j = 0, k = 0; j < nb_used; j++) {
                    if (data_blocs[j] != NULL)
                        new_data_blocs[k++] = data_blocs[j];
                }
            } else {
                new_data_blocs = NULL;
            }
            /* Free bloc list */
            vfree(dev_list[i].btype);
            device->btype = NULL;
            vfree(data_blocs);
            data_blocs = new_data_blocs;

            kfree(dev_list);
            dev_list = new_dev_list;
            nb_used -= device->nb_mapped;
            nb_allocs -= device->nb_blocs;
            nb_remaps--;
            unlock_all();
            REMAP_DEBUG("Unmap done\n");
            
            return 0;
        }
    }
    REMAP_ERROR("Device not found (%x)\n", rdev);
    unlock_all();

    return -EINVAL;

unmap_error:
    if (device != NULL) {
        vfree(device->btype);
        device->btype = NULL;
    }
    vfree(new_data_blocs);
    kfree(new_dev_list);
    unlock_all();

    return ret;
}

static int remap_getmap (unsigned char *addr)
{
    user_remap_t user_remap;
    rem_dev_t *device;
    kdev_t rdev;
    int i;
    int ret;

    lock_all();
    copy_from_user(&user_remap, addr, sizeof(user_remap_t));
    rdev = user_remap.dev;
    memset(&user_remap, 0, sizeof(user_remap_t));
    /* Check that we didn't already remapped this device */
    for (i = 0; dev_list != NULL && i < nb_remaps; i++) {
        device = &dev_list[i];
        if (dev_list[i].device == rdev) {
            user_remap.dev = MAKEDEV(REMAP_MAJOR, i);
            ret = 0;
            goto end_getmap;
        }
    }
    ret = -ENODEV;

 end_getmap:
    unlock_all();
    copy_to_user(addr, &user_remap, sizeof(user_remap_t));

    return ret;
}

static int remap_chr_ioctl(struct inode *inode, struct file *file,
                           uint command, ulong a)
{
    unsigned char *addr;

    addr = (unsigned char *)a;
    switch (command) {
    case REMAP_DOMAP:
        REMAP_DEBUG("DOMAP\n");
        return remap_domap(addr);
    case REMAP_DOUNMAP:
        REMAP_DEBUG("DOUNMAP\n");
        return remap_dounmap(addr);
    case REMAP_GETMAP:
        REMAP_DEBUG("GETMAP\n");
        return remap_getmap(addr);
    default:
        REMAP_ERROR("UNKNOWN IOCTL %d\n", command);
        break;
    }

    return -EINVAL;
}

static struct block_device_operations remap_blk_fops =
{
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 14))
	owner:		THIS_MODULE,
#endif /* (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 14)) */
	open: 		remap_blk_open,
	release:	remap_blk_close,
};

static struct file_operations remap_chr_fops = {
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 14))
	owner:		THIS_MODULE,
#endif /* (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 14)) */
	ioctl:		remap_chr_ioctl,
	open:		remap_chr_open,
	release:	remap_chr_close,
};

static int init_remap (void)
{
    request_queue_t *q;

    lock_all();

    /* Register into devfs */
#if LINUX_VERSION_CODE < 0x020300 || defined(USE_NO_DEVFS)
    if (register_blkdev(REMAP_MAJOR, "bdremap", &remap_blk_fops))
#else
    if (devfs_register_blkdev(REMAP_MAJOR, "bdremap", &remap_blk_fops))
#endif
    {
        unlock_all();
        return -EIO;
    }
#if LINUX_VERSION_CODE >= 0x020300 && !defined(USE_NO_DEVFS)
    devfs_handle = devfs_mk_dir(NULL, "bdremap", NULL);
    devfs_register_series(devfs_handle, "%u", REMAP_MAX_MINORS,
                          DEVFS_FL_DEFAULT, REMAP_MAJOR, 0,
                          S_IFBLK | S_IRUSR | S_IWUSR | S_IRGRP,
                          &remap_blk_fops, NULL);
#endif
    /* Register our make request function */
    q = blk_get_queue(MAKEDEV(REMAP_MAJOR, 0));
    blk_queue_make_request(BLK_DEFAULT_QUEUE(REMAP_MAJOR), remap_bloc_request);
#if LINUX_VERSION_CODE < 0x020300 || defined(USE_NO_DEVFS)
    register_chrdev(REMAP_MAJOR, "bdremap", &remap_chr_fops);
#else
    devfs_register_chrdev(REMAP_MAJOR, "bdremap", &remap_chr_fops);
    devfs_register(devfs_handle, "ctrl", DEVFS_FL_DEFAULT, REMAP_MAJOR, 0,
                   S_IFCHR | S_IRUSR | S_IWUSR | S_IRGRP,
                   &remap_chr_fops, NULL);
#endif
    unlock_all();

    return 0;
}

static void cleanup_remap (void)
{
    rem_dev_t *device;
    unsigned short i;

    lock_all();
    /* Free all remapped devices */
    if (dev_list != NULL) {
        for (i = 0; i < nb_remaps; i++) {
            REMAP_DEBUG("Free device %d\n", i);
            device = &dev_list[i]; 
            if (device->btype != NULL)
                vfree(device->btype);
        }
        REMAP_DEBUG("Free device list\n");
        kfree(dev_list);
    }
    dev_list = NULL;
    nb_remaps = 0;

    /* Free all data blocs */
    REMAP_DEBUG("Free data blocs\n");
    if (data_blocs != NULL) {
        for (i = 0; i < nb_used; i++)
            vfree(data_blocs[i]);
        REMAP_DEBUG("Free special data blocs\n");
        vfree(data_blocs);
    }
    data_blocs = NULL;
    nb_used = 0;
    nb_allocs = 0;
    
    REMAP_DEBUG("Unregister blkdev\n");
#if LINUX_VERSION_CODE < 0x020300 || defined(USE_NO_DEVFS)
    unregister_chrdev(REMAP_MAJOR, "bdremap");
    unregister_blkdev(REMAP_MAJOR, "bdremap");
#else
    devfs_unregister_chrdev(REMAP_MAJOR, "bdremap");
    devfs_unregister_blkdev(REMAP_MAJOR, "bdremap");
    if (devfs_handle != NULL)
        devfs_unregister(devfs_handle);
#endif
    unlock_all();
}

module_init(init_remap);
module_exit(cleanup_remap);

/** Module description
 */
MODULE_AUTHOR("J. Mayer <l_indien@magic.fr>");
MODULE_DESCRIPTION("Block device remapper");
MODULE_SUPPORTED_DEVICE("remap");
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 10))
MODULE_LICENSE("GPL");
#endif
