/*
 * lib_remap.c
 *
 * This library intend to contain all routines needed for a program
 * to control bd_remap module.
 */
/*
 * Copyright (C) 2003-2004 by Jocelyn Mayer <l_indien@magic.fr>
 * distributed under the GNU Lesser General Public License
 *
 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.
 */
/* History:
 * 2003-02-05: Created
 */

#define _GNU_SOURCE
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

#include <sys/stat.h>
#include <sys/mount.h>
#include <sys/ioctl.h>
#include <dirent.h>

#include "bd_remap.h"
#include "lib_remap.h"

#ifndef FIBMAP
#define FIBMAP     _IO(0x00,1)  /* bmap access */
#endif
#ifndef FIGETBSZ
#define FIGETBSZ   _IO(0x00,2)  /* get the block size used for bmap */
#endif

static FILE *msg_stream, *error_stream;

void bdrm_register_msg_stream (FILE * stream)
{
    msg_stream = stream;
}

void bdrm_restore_msg_stream (void)
{
    msg_stream = stdout;
}

void bdrm_register_error_stream (FILE * stream)
{
    error_stream = stream;
}

void bdrm_restore_error_stream (void)
{
    error_stream = stderr;
}

__attribute__ (( format(printf,1,2) ))
static int bdrm_do_message (const char *format, ...)
{
    va_list ap;

    if (msg_stream == NULL)
        return 0;
    va_start(ap, format);
    vfprintf(msg_stream, format, ap);
    va_end(ap);

    return 0;
}

__attribute__ (( format(printf,1,2) ))
static int bdrm_do_error (const char *format, ...)
{
    va_list ap;

    if (error_stream == NULL)
        return 0;
    va_start(ap, format);
    vfprintf(error_stream, format, ap);
    va_end(ap);

    return 0;
}

/* Dump the type of remapping for a bloc or a file */
static void bdrm_dump_type (int type)
{
    switch (type) {
    case REMAP_DISCARD:
        bdrm_do_message("discarded\n");
        break;
    case REMAP_EMPTY:
        bdrm_do_message("empty\n");
        break;
    case REMAP_EMPTYFF:
        bdrm_do_message("empty (0xFF)\n");
        break;
    case REMAP_NOREMAP:
        bdrm_do_message("not remapped\n");
        break;
    case REMAP_COPY:
        bdrm_do_message("copied\n");
        break;
    default:
        bdrm_do_message("unknown type: %d\n", type);
    }
 }

/* Dump the whole state before sending request to the kernel */
static void bdrm_dump_all (const char *const mp,
                           const user_remap_t *const remap,
                           const file_t *const file_list,
                           const unsigned int nb_files,
                           const int *const btype)
{
    unsigned long st_bloc, end_bloc;
    unsigned int len;
    unsigned int i;
    int type;

    bdrm_do_message("Mount point: %s %d %d %lu\n", mp, remap->dev >> 8,
           remap->dev & 0xFF, remap->nb_blocs);
    bdrm_do_message("Special files:\n");
    for (i = 0; i < nb_files; i++) {
        bdrm_do_message("* %s", file_list[i].name);
        len = strlen(file_list[i].name) + 2;
        while (len < 40) {
            bdrm_do_message("\t");
            len += 8;
        }
        if ((file_list[i].type & 0x100) == 0) {
            bdrm_do_message("skipped\n");
        } else {
            bdrm_dump_type(file_list[i].type);
        }
    }
    if (btype != NULL) {
        type = REMAP_INIT;
        st_bloc = 0;
        end_bloc = 0;
        for (i = 0; i < remap->nb_blocs; i++) {
            if (type != btype[i]) {
                if (type != REMAP_INIT) {
                    end_bloc = i - 1;
                    bdrm_do_message("Bloc %lu to %lu: ", st_bloc, end_bloc);
                    bdrm_dump_type(type);
                }
                type = btype[i];
                st_bloc = i;
            }
        }
        if (type != REMAP_INIT) {
            end_bloc = i - 1;
            bdrm_do_message("Bloc %lu to %lu: ", st_bloc, end_bloc);
            bdrm_dump_type(type);
        }
    }
}

/* Make a filename: may be absolute or composed from a base */
static char *bdrm_make_fname (const char *base, const char *file)
{
    char *fname, *pos;
    unsigned int blen, flen;

    blen = strlen(base);
    flen = strlen(file);
    fname = malloc(blen + flen + 2);
    if (fname == NULL) {
        bdrm_do_error("Memory error\n");
        return NULL;
    }
    memset(fname, 0, blen + flen + 2);
    memcpy(fname, base, blen);
    if (blen > 0) {
        pos = fname + blen - 1;
        if (*pos != '/') {
            if (*file != '/')
                *pos++ = '/';
        } else {
            if (*file == '/')
                file++;
        }
        pos++;
    } else {
        pos = fname;
    }
    memcpy(pos, file, flen);
    pos[flen + 1] = '\0';
    if (access(fname, R_OK) < 0) {
        if (*file != '/') {
            *fname = '/';
            memcpy(fname + 1, file, flen);
            fname[flen + 1] = '\0';
        } else {
            memcpy(fname, file, flen);
            fname[flen] = '\0';
        }
        if (access(fname, R_OK) < 0) {
            free(fname);
            return NULL;
        }
    }

    return fname;
}

/* Add a file in the list a file not to be remapped */
int bdrm_add_file (file_t **const file_list,
                   unsigned int *const nb_files,
                   const char *name, const int type, const int verbose)
{
    file_t *new_list;

    new_list = realloc(*file_list, (*nb_files + 1) * sizeof(file_t));
    if (new_list == NULL) {
        bdrm_do_error("Memory error\n");
        return -1;
    }
    *file_list = new_list;
    new_list[*nb_files].name = strdup(name);
    if (new_list[*nb_files].name == NULL) {
        bdrm_do_error("Memory error\n");
        return -1;
    }
    new_list[*nb_files].type = type;
    new_list[*nb_files].blocs = NULL;
    if (verbose != 0) {
        bdrm_do_message("Add file %s with type ", name);
        bdrm_dump_type(type);
    }
    (*nb_files)++;

    return 0;
}

void bdrm_free_list (file_t **file_list, unsigned int nb_files)
{
    file_t *file;
    unsigned int i;

    for (i = 0 ; i < nb_files; i++) {
        file = &(*file_list)[i];
        free(file->name);
        free(file->blocs);
    }
    free(*file_list);
    *file_list = NULL;
}

/* Simplistic seekdir for use with libc's that don't have one */
static struct dirent *bdrm_do_seekdir (DIR *pdir, const char *name)
{
    struct dirent *entry;

    if (pdir == NULL || name == NULL)
        return NULL;

    while ((entry = readdir(pdir)) != NULL) {
        if (strcmp(entry->d_name, name) == 0)
            break;
    }

    return entry;
}

/* Add files recursively */
static void bdrm_add_recursive (file_t **const file_list,
                                unsigned int *const nb_files,
                                const char *name, int type,
                                unsigned int device, const int verbose)
{
    struct stat st;
    char filename[PATH_MAX];
    DIR *pdir;
    struct dirent *entry;
    unsigned int nlen, flen;

    pdir = opendir(name);
    if (pdir == NULL)
        return;

    nlen = strlen(name);
    memcpy(filename, name, nlen);
    if (filename[nlen - 1] != '/')
        filename[nlen++] = '/';
    while ((entry = readdir(pdir)) != NULL) {
        if (entry->d_name[0] == '.' &&
            (entry->d_name[1] == '\0' ||
             (entry->d_name[1] == '.' && entry->d_name[2] == '\0')))
            continue;
        flen = strlen(entry->d_name);
        memcpy(filename + nlen, entry->d_name, flen);
        filename[nlen + flen] = '\0';
        if (lstat(filename, &st) < 0)
            continue;
        if (bdrm_add_file(file_list, nb_files, filename, type, verbose) < 0)
            continue;
        if (S_ISDIR(st.st_mode)) {
            if (st.st_dev != device)
                continue;
            closedir(pdir);
            bdrm_add_recursive(file_list, nb_files, filename,
                          type, device, verbose);
            pdir = opendir(name);
            if (pdir == NULL)
                break;
            entry = bdrm_do_seekdir(pdir, filename);
            if (entry == NULL)
                break;
        }
    }
}

int bdrm_do_unmap (const unsigned char *chrdev,
                   const int dev, const int simulate)
{
    user_remap_t remap;
    int fd;
    int ret;

    if (chrdev == NULL)
        chrdev = "/dev/remap";
    /* Unmap request */
    memset(&remap, 0, sizeof(user_remap_t));
    remap.dev = dev;
    fd = open(chrdev, O_RDONLY);
    if (fd < 0) {
        bdrm_do_error("Unable to open %s : %s\n", chrdev, strerror(errno));
        return -1;
    }
    ret = 0;
    if (simulate == 0) {
        if (ioctl(fd, REMAP_DOUNMAP, remap) < 0) {
            bdrm_do_error("Unable to unmap dev %d : %s\n", dev, strerror(errno));
            ret = -1;
        }
    }
    close(fd);

    return ret;
}

int bdrm_do_map (const unsigned char *chrdev, file_t **const file_list,
                 unsigned int *const nb_files, const char *const mp,
                 const unsigned int blocsize, const int verbose,
                 const int simulate, int *minor)
{
    struct stat st;
    unsigned char name[16];
    user_remap_t remap;
    unsigned int *bcount;
    unsigned char *file_name;
    unsigned long bloc, rbloc;
    unsigned int __blocsize, blocdec, ratio;
    int *btype;
    int inv;
    int devsize;
    int fd;
    unsigned int i, j;
    int ret = -1;

    if (chrdev == NULL)
        chrdev = "/dev/remap";
    if (lstat(mp, &st) < 0) {
        bdrm_do_error("can't stat: %s\n", mp);
        return -1;
    }
    snprintf(name, sizeof(name), "/tmp/__%02x__%02x__",
            (unsigned int)st.st_dev >> 8, (unsigned int)st.st_dev & 0xFF);
    unlink(name);
    if (mknod(name, S_IFBLK | 0755, st.st_dev) < 0 && errno != EEXIST) {
        bdrm_do_error("can't create: %s\n", name);
        return -1;
    }
    fd = open(name, O_RDONLY);
    if (fd < 0) {
        bdrm_do_error("can't open: %s\n", name);
        return -1;
    }
    btype = NULL;
    bcount = NULL;
    if (ioctl(fd, BLKGETSIZE, &devsize) < 0) {
        bdrm_do_error("can't get size of %s : %s\n", name, strerror(errno));
        goto end;
    }
    close(fd);
    unlink(name);
    blocdec = 0;
    __blocsize = blocsize;
    while (__blocsize > REMAP_BLOC_SIZE) {
        blocdec++;
        __blocsize /= 2;
    }
    bloc = (devsize * 512) / (REMAP_BLOC_SIZE << blocdec);
    remap.dev = st.st_dev;
    remap.nb_blocs = bloc;
    remap.bloc_dec = blocdec;
    btype = remap.type;
    bcount = malloc(sizeof(int) * bloc);
    if (bcount == NULL) {
        bdrm_do_error("Memory error\n");
        goto end;
    }
    memset(bcount, 0, sizeof(int) * bloc);
    if ((unsigned long)st.st_blksize <= (REMAP_BLOC_SIZE << blocdec)) {
        ratio = (REMAP_BLOC_SIZE << blocdec) / st.st_blksize;
        inv = 1;
    } else {
        ratio = st.st_blksize / (REMAP_BLOC_SIZE << blocdec);
        inv = 0;
    }

    btype = malloc(sizeof(int) * bloc);
    if (btype == NULL) {
        bdrm_do_error("Memory error\n");
        goto end;
    }
    /* Pre-initialise all blocs as copied */
    for (i = 0; i < remap.nb_blocs; i++)
        btype[i] = REMAP_COPY;

    for (i = 0; i < *nb_files; i++) {
        file_name = bdrm_make_fname(mp, (*file_list)[i].name);
        if (file_name == NULL) {
            bdrm_do_error("WARNING: unknown file: %s - Skip it\n",
                     (*file_list)[i].name);
            continue;
        }
        fd = open(file_name, O_RDONLY);
        if (fd < 0) {
            bdrm_do_error("WARNING: can't open file: %s for reading - "
                     "Skip it\n", (*file_list)[i].name);
            continue;
        }
        if (fstat(fd, &st) < 0) {
            bdrm_do_error("WARNING: can't stat %s - Skip it\n",
                    (*file_list)[i].name);
            close(fd);
            continue;
        }
        if (st.st_dev != remap.dev) {
            bdrm_do_error("WARNING: file %s is on another device - "
                    "Skip it\n", (*file_list)[i].name);
            close(fd);
            continue;
        }
        if (S_ISDIR(st.st_mode)) {
            if (verbose != 0) {
                bdrm_do_message("Add directory %s content recursively\n",
                           (*file_list)[i].name);
            }
            bdrm_add_recursive(file_list, nb_files, file_name,
                               (*file_list)[i].type, remap.dev, verbose);
        } else {
            __blocsize = st.st_blksize;
            for (j = 0; j < (st.st_size + __blocsize - 1) / __blocsize; j++) {
                bloc = j;
                if (ioctl(fd, FIBMAP, &bloc) < 0) {
                    bdrm_do_error("Error : can't get block for file: %s\n",
                            (*file_list)[i].name);
                    close(fd);
                    goto end;
                }
                /* Skip zero blocs:
                 * these one can appear on compressed filesystems
                 */
                if (bloc == 0)
                    continue;
                if (inv == 1) {
                    rbloc = bloc / ratio;
                    if ((*file_list)[i].type != btype[rbloc])
                        bcount[rbloc] = 0;
                    bcount[rbloc]++;
#if 0
                    if (verbose > 2) {
                        bdrm_do_message("Set bloc %lu for file %s "
                                   "to type %d (%d - %d)\n",
                                   rbloc, file_name, (*file_list)[i].type,
                                   btype[rbloc], (1 << blocdec) + 1);
                    }
#endif
                    btype[rbloc] = (*file_list)[i].type;
                } else {
                    for (rbloc = bloc * ratio;
                         rbloc < ((bloc + 1) * ratio); rbloc++) {
#if 0
                        if (verbose > 2) {
                            bdrm_do_message("Set bloc %lu for file %s "
                                       "to type %d (%d - %d)\n",
                                       rbloc, file_name, (*file_list)[i].type,
                                       btype[rbloc], (1 << blocdec) + 1);
                        }
#endif
                        btype[rbloc] = (*file_list)[i].type;
                    }
                }
            }
        }
        (*file_list)[i].type |= 0x100;
        close(fd);
        free(file_name);
    }
    if (inv == 1) {
        /* If many device blocs can be remapped in the same bloc in memory,
         * always copy it if one or more subpart of the memory bloc need
         * to be copied
         */
        for (bloc = 0; bloc < remap.nb_blocs; bloc++) {
            if (bcount[bloc] != ratio)
                btype[bloc] = REMAP_COPY;
        }
    }
    bdrm_dump_all(mp, &remap, *file_list, *nb_files, btype);
    fd = open(chrdev, O_RDONLY);
    if (fd < 0) {
        bdrm_do_error("Unable to open device: /dev/remap\n");
        goto end;
    }
    if (simulate == 0) {
        if (ioctl(fd, REMAP_DOMAP, &remap) < 0) {
            bdrm_do_error("map failed for device: %s\n", mp);
            close(fd);
            goto end;
        }
        *minor = remap.dev & 0xFF;
    }
    close(fd);
    ret = 0;

end:
    free(btype);
    free(bcount);

    return ret;
}

int bdrm_get_map (const unsigned char *chrdev, const char *const mp,
                  int *minor)
{
    struct stat st;
    user_remap_t remap;
    int fd;
    int ret;

    if (chrdev == NULL)
        chrdev = "/dev/remap";
    if (lstat(mp, &st) < 0) {
        bdrm_do_error("can't stat: %s\n", mp);
        return -1;
    }
    memset(&remap, 0, sizeof(user_remap_t));
    remap.dev = st.st_dev;
    fd = open(chrdev, O_RDONLY);
    if (fd < 0) {
        bdrm_do_error("Unable to open %s : %s\n", chrdev, strerror(errno));
        return -1;
    }
    if (ioctl(fd, REMAP_GETMAP, &remap) < 0) {
        ret = -1;
        *minor = -1;
    } else {
        ret = 0;
        *minor = remap.dev & 0xFF;
    }
    close(fd);

    return ret;
}
